# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from typing import Union
import json
import os
import shutil
import sys
import urllib.request
from subprocess import DEVNULL, STDOUT, call
if os.name == 'nt':
    from ctypes import wintypes
    from ctypes import WinDLL as DllLoader
else:
    from ctypes import CDLL as DllLoader
import numpy as np
from casadi import DM, MX, SX, CasadiMeta, Function

ALLOWED_CASADI_VERSIONS = (
    '3.4.0',
    '3.4.5',
    '3.5.1',
    '3.5.2',
    '3.5.3',
    '3.5.4',
    '3.5.6',
    '3.5.5',
    '3.6.0',
    '3.6.1',
    '3.6.2',
    '3.6.3',
    '3.6.4',
    '3.6.5',
    '3.6.6',
    '3.6.7',
)

TERA_VERSION = "0.0.34"

PLATFORM2TERA = {
    "linux": "linux",
    "darwin": "osx",
    "win32": "windows"
}

ACADOS_INFTY = 1e10

def check_if_square(mat: np.ndarray, name: str):
    if mat.shape[0] != mat.shape[1]:
        raise ValueError(f"Matrix {name} must be square, got shape {mat.shape}.")
    return

def get_acados_path():
    ACADOS_PATH = os.environ.get('ACADOS_SOURCE_DIR')
    if not ACADOS_PATH:
        acados_template_path = os.path.dirname(os.path.abspath(__file__))
        acados_path = os.path.join(acados_template_path, '..','..','..')
        ACADOS_PATH = os.path.realpath(acados_path)
        msg = 'Warning: Did not find environment variable ACADOS_SOURCE_DIR, '
        msg += 'guessed ACADOS_PATH to be {}.\n'.format(ACADOS_PATH)
        msg += 'Please export ACADOS_SOURCE_DIR to avoid this warning.'
        print(msg)
    return ACADOS_PATH


def get_python_interface_path():
    ACADOS_PYTHON_INTERFACE_PATH = os.environ.get('ACADOS_PYTHON_INTERFACE_PATH')
    if not ACADOS_PYTHON_INTERFACE_PATH:
        acados_path = get_acados_path()
        ACADOS_PYTHON_INTERFACE_PATH = os.path.join(acados_path, 'interfaces', 'acados_template', 'acados_template')
    return ACADOS_PYTHON_INTERFACE_PATH


def get_tera_exec_path():
    TERA_PATH = os.environ.get('TERA_PATH')
    if not TERA_PATH:
        TERA_PATH = os.path.join(get_acados_path(), 'bin', 't_renderer')
        if os.name == 'nt':
            TERA_PATH += '.exe'
    return TERA_PATH


def acados_lib_is_compiled_with_openmp(acados_lib: DllLoader, verbose: bool) -> bool:
    # find out if acados was compiled with OpenMP
    try:
        acados_lib_uses_omp = getattr(acados_lib, 'omp_get_thread_num') is not None
    except AttributeError as e:
        acados_lib_uses_omp = False
    if verbose:
        if acados_lib_uses_omp:
            print('acados was compiled with OpenMP.')
        else:
            print('acados was compiled without OpenMP.')
    return acados_lib_uses_omp


def get_shared_lib(shared_lib_name: str, winmode = None) -> DllLoader:
    if winmode is not None:
        shared_lib = DllLoader(shared_lib_name, winmode=winmode)
    else:
        # for compatibility with older python versions
        shared_lib = DllLoader(shared_lib_name)
    return shared_lib


def check_casadi_version():
    casadi_version = CasadiMeta.version()
    if casadi_version in ALLOWED_CASADI_VERSIONS:
        return
    else:
        msg =  'Warning: Please note that the following versions of CasADi are '
        msg += 'officially supported: {}.\n '.format(" or ".join(ALLOWED_CASADI_VERSIONS))
        msg += 'If there is an incompatibility with the CasADi generated code, '
        msg += 'please consider changing your CasADi version.\n'
        msg += 'Version {} currently in use.'.format(casadi_version)
        print(msg)

def check_casadi_version_supports_p_global():
    try:
        from casadi import extract_parametric, cse
    except:
        raise Exception("CasADi version does not support extract_parametric or cse functions.\nNeeds nightly-se2 release or later, see: https://github.com/casadi/casadi/releases/tag/nightly-se2")


def get_simulink_default_opts():
    python_interface_path = get_python_interface_path()
    abs_path = os.path.join(python_interface_path, 'simulink_default_opts.json')
    with open(abs_path , 'r') as f:
        simulink_default_opts = json.load(f)
    return simulink_default_opts


def is_column(x):
    if isinstance(x, np.ndarray):
        if x.ndim == 1:
            return True
        elif x.ndim == 2 and x.shape[1] == 1:
            return True
        else:
            return False
    elif isinstance(x, (MX, SX, DM)):
        if x.shape[1] == 1:
            return True
        elif x.shape[0] == 0 and x.shape[1] == 0:
            return True
        else:
            return False
    elif x == None or x == []:
        return False
    else:
        raise Exception("is_column expects one of the following types: np.ndarray, casadi.MX, casadi.SX."
                        + " Got: " + str(type(x)))


def is_empty(x):
    if isinstance(x, (MX, SX, DM)):
        return x.is_empty()
    elif isinstance(x, np.ndarray):
        return True if np.prod(x.shape) == 0 else False
    elif x is None:
        return True
    elif isinstance(x, (set, list)):
        return True if len(x) == 0 else False
    elif isinstance(x, (float, int)):
        return False
    else:
        raise Exception("is_empty expects one of the following types: casadi.MX, casadi.SX, "
                        + "None, numpy array empty list, set. Got: " + str(type(x)))


def casadi_length(x):
    if isinstance(x, (MX, SX, DM)):
        return int(np.prod(x.shape))
    elif x is None:
        return 0
    elif isinstance(x, list):
        return len(x)
    else:
        raise Exception("casadi_length expects one of the following types: casadi.MX, casadi.SX."
                        + " Got: " + str(type(x)))

def get_shared_lib_ext():
    if sys.platform == 'darwin':
        return '.dylib'
    elif os.name == 'nt':
        return '.dll'
    else:
        return '.so'

def get_shared_lib_dir():
    if os.name == 'nt':
        return 'bin'
    else:
        return 'lib'

def get_shared_lib_prefix():
    if os.name == 'nt':
        return ''
    else:
        return 'lib'

def get_tera() -> str:
    tera_path = get_tera_exec_path()
    acados_path = get_acados_path()

    # check if tera exists and is executable
    if os.path.exists(tera_path) and os.access(tera_path, os.X_OK):
        return tera_path

    repo_url = "https://github.com/acados/tera_renderer/releases"
    url = "{}/download/v{}/t_renderer-v{}-{}".format(
        repo_url, TERA_VERSION, TERA_VERSION, PLATFORM2TERA[sys.platform])

    manual_install = 'For manual installation follow these instructions:\n'
    manual_install += '1 Download binaries from {}\n'.format(url)
    manual_install += '2 Copy them in {}/bin\n'.format(acados_path)
    manual_install += '3 Strip the version and platform from the binaries: '
    manual_install += 'as t_renderer-v0.0.34-X -> t_renderer)\n'
    manual_install += '4 Enable execution privilege on the file "t_renderer" with:\n'
    manual_install += '"chmod +x {}"\n\n'.format(tera_path)

    msg = "\n"
    msg += 'Tera template render executable not found, '
    msg += 'while looking in path:\n{}\n'.format(tera_path)
    msg += 'In order to be able to render the templates, '
    msg += 'you need to download the tera renderer binaries from:\n'
    msg += '{}\n\n'.format(repo_url)
    msg += 'Do you wish to set up Tera renderer automatically?\n'
    msg += 'y/N? (press y to download tera or any key for manual installation)\n'

    if input(msg) != 'y':
        msg_cancel = "\nYou cancelled automatic download.\n\n"
        msg_cancel += manual_install
        msg_cancel += "Once installed re-run your script.\n\n"
        print(msg_cancel)

        sys.exit(1)

    # check if parent directory exists otherwise create it
    tera_dir = os.path.split(tera_path)[0]
    if not os.path.exists(tera_dir):
        print(f"Creating directory {tera_dir}")
        os.makedirs(tera_dir)

    # Download tera
    print(f"Dowloading {url}")
    with urllib.request.urlopen(url) as response, open(tera_path, 'wb') as out_file:
        shutil.copyfileobj(response, out_file)
    print("Successfully downloaded t_renderer.")
    # make executable
    os.chmod(tera_path, 0o755)
    print("Successfully made t_renderer executable.")
    return tera_path


def render_template(in_file, out_file, output_dir, json_path, template_glob=None):

    acados_path = os.path.dirname(os.path.abspath(__file__))
    if template_glob is None:
        head, in_file = os.path.split(in_file)
        template_glob = os.path.join(acados_path, 'c_templates_tera', head, '**', '*')
    cwd = os.getcwd()

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    os.chdir(output_dir)

    tera_path = get_tera()

    # call tera as system cmd
    os_cmd = f"{tera_path} '{template_glob}' '{in_file}' '{json_path}' '{out_file}'"
    # Windows cmd.exe can not cope with '...', so use "..." instead:
    if os.name == 'nt':
        os_cmd = os_cmd.replace('\'', '\"')

    status = os.system(os_cmd)
    if status != 0:
        raise Exception(f'Rendering of {in_file} failed!\n\nAttempted to execute OS command:\n{os_cmd}\n\n')

    os.chdir(cwd)


def casadi_expr_to_string(expr) -> str:
    string = ''
    for ii in range(casadi_length(expr)):
        string += f"{expr[ii,:]}\n"
    return string

## Conversion functions
def make_object_json_dumpable(input):
    if isinstance(input, (np.ndarray)):
        return input.tolist()
    elif isinstance(input, (SX)):
        try:
            return input.serialize()
            # for more readable json output:
            # return casadi_expr_to_string(input)
        except: # for older CasADi versions
            return ''
    elif isinstance(input, (MX)):
        # NOTE: MX expressions can not be serialized, only Functions.
        return input.__str__()
    elif isinstance(input, (DM)):
        return input.full()
    else:
        raise TypeError(f"Cannot make input of type {type(input)} dumpable.")


def format_class_dict(d):
    """
    removes the __ artifact from class to dict conversion
    """
    out = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = format_class_dict(v)

        out_key = k.split('__', 1)[-1]
        out[k.replace(k, out_key)] = v
    return out

def get_default_simulink_opts() -> dict:
    print("get_default_simulink_opts is deprecated, use get_simulink_default_opts instead."
          + " This function will be removed in a future release.")
    return get_simulink_default_opts()


def J_to_idx(J):
    if not isinstance(J, np.ndarray):
        raise Exception('J_to_idx: J must be a numpy array.')
    if J.ndim != 2:
        raise Exception('J_to_idx: J must be a 2D numpy array.')
    nrows = J.shape[0]
    idx = np.zeros((nrows, ))
    for i in range(nrows):
        this_idx = np.nonzero(J[i,:])[0]
        if len(this_idx) != 1:
            raise Exception('Invalid J matrix structure detected, ' \
                'must contain exactly one nonzero element per row.')
        if this_idx.size > 0 and J[i,this_idx[0]] != 1:
            raise Exception('J matrices can only contain 1 and 0 entries.')
        idx[i] = this_idx[0]
    return idx


def J_to_idx_slack(J):
    nrows = J.shape[0]
    ncol = J.shape[1]
    idx = np.zeros((ncol, ))
    i_idx = 0
    for i in range(nrows):
        this_idx = np.nonzero(J[i,:])[0]
        if len(this_idx) == 1:
            idx[i_idx] = i
            i_idx = i_idx + 1
        elif len(this_idx) > 1:
            raise Exception('J_to_idx_slack: Invalid J matrix. ' \
                'Found more than one nonzero in row ' + str(i))
        if this_idx.size > 0 and J[i,this_idx[0]] != 1:
            raise Exception('J_to_idx_slack: J matrices can only contain 1s, ' \
                 'got J(' + str(i) + ', ' + str(this_idx[0]) + ') = ' + str(J[i,this_idx[0]]) )
    if not i_idx == ncol:
            raise Exception('J_to_idx_slack: J must contain a 1 in every column!')
    return idx


def check_if_nparray_and_flatten(val, name) -> np.ndarray:
    if not isinstance(val, np.ndarray):
        raise Exception(f"{name} must be a numpy array, got {type(val)}")
    return val.reshape(-1)


def check_if_2d_nparray(val, name) -> None:
    if not isinstance(val, np.ndarray):
        raise Exception(f"{name} must be a numpy array, got {type(val)}")
    if val.ndim != 2:
        raise Exception(f"{name} must be a 2D numpy array, got shape {val.shape}")
    return

def print_J_to_idx_note():
    print("NOTE: J* matrix is converted to zero based vector idx* vector, which is returned here.")


def acados_dae_model_json_dump(model):

    # load model
    x = model.x
    xdot = model.xdot
    u = model.u
    z = model.z
    p = model.p

    f_impl = model.f_impl_expr
    model_name = model.name

    # create struct with impl_dae_fun, casadi_version
    fun_name = model_name + '_impl_dae_fun'
    impl_dae_fun = Function(fun_name, [x, xdot, u, z, p], [f_impl])

    casadi_version = CasadiMeta.version()
    str_impl_dae_fun = impl_dae_fun.serialize()

    dae_dict = {"str_impl_dae_fun": str_impl_dae_fun, "casadi_version": casadi_version}

    # dump
    json_file = model_name + '_acados_dae.json'
    with open(json_file, 'w') as f:
        json.dump(dae_dict, f, default=make_object_json_dumpable, indent=4, sort_keys=True)
    print("dumped ", model_name, " dae to file:", json_file, "\n")


def set_up_imported_gnsf_model(acados_ocp):

    gnsf = acados_ocp.gnsf_model

    # load model
    phi_fun = Function.deserialize(gnsf['phi_fun'])
    phi_fun_jac_y = Function.deserialize(gnsf['phi_fun_jac_y'])
    phi_jac_y_uhat = Function.deserialize(gnsf['phi_jac_y_uhat'])
    get_matrices_fun = Function.deserialize(gnsf['get_matrices_fun'])

    # obtain gnsf dimensions
    size_gnsf_A = get_matrices_fun.size_out(0)
    acados_ocp.dims.gnsf_nx1 = size_gnsf_A[1]
    acados_ocp.dims.gnsf_nz1 = size_gnsf_A[0] - size_gnsf_A[1]
    acados_ocp.dims.gnsf_nuhat = max(phi_fun.size_in(1))
    acados_ocp.dims.gnsf_ny = max(phi_fun.size_in(0))
    acados_ocp.dims.gnsf_nout = max(phi_fun.size_out(0))

    # save gnsf functions in model
    acados_ocp.model.phi_fun = phi_fun
    acados_ocp.model.phi_fun_jac_y = phi_fun_jac_y
    acados_ocp.model.phi_jac_y_uhat = phi_jac_y_uhat
    acados_ocp.model.get_matrices_fun = get_matrices_fun

    # get_matrices_fun = Function([model_name,'_gnsf_get_matrices_fun'], {dummy},...
    #  {A, B, C, E, L_x, L_xdot, L_z, L_u, A_LO, c, E_LO, B_LO,...
    #   nontrivial_f_LO, purely_linear, ipiv_x, ipiv_z, c_LO});
    get_matrices_out = get_matrices_fun(0)
    acados_ocp.model.gnsf_nontrivial_f_LO = int(get_matrices_out[12])
    acados_ocp.model.gnsf_purely_linear = int(get_matrices_out[13])

    if "f_lo_fun_jac_x1k1uz" in gnsf:
        f_lo_fun_jac_x1k1uz = Function.deserialize(gnsf['f_lo_fun_jac_x1k1uz'])
        acados_ocp.model.f_lo_fun_jac_x1k1uz = f_lo_fun_jac_x1k1uz
    else:
        dummy_var_x1 = SX.sym('dummy_var_x1', acados_ocp.dims.gnsf_nx1)
        dummy_var_x1dot = SX.sym('dummy_var_x1dot', acados_ocp.dims.gnsf_nx1)
        dummy_var_z1 = SX.sym('dummy_var_z1', acados_ocp.dims.gnsf_nz1)
        dummy_var_u = SX.sym('dummy_var_z1', acados_ocp.dims.nu)
        dummy_var_p = SX.sym('dummy_var_z1', acados_ocp.dims.np)
        empty_var = SX.sym('empty_var', 0, 0)

        empty_fun = Function('empty_fun', \
            [dummy_var_x1, dummy_var_x1dot, dummy_var_z1, dummy_var_u, dummy_var_p],
                [empty_var])
        acados_ocp.model.f_lo_fun_jac_x1k1uz = empty_fun

    del acados_ocp.gnsf_model


def idx_perm_to_ipiv(idx_perm):
    n = len(idx_perm)
    vec = list(range(n))
    ipiv = np.zeros(n)

    print(n, idx_perm)
    # import pdb; pdb.set_trace()
    for ii in range(n):
        idx0 = idx_perm[ii]
        for jj in range(ii,n):
            if vec[jj]==idx0:
                idx1 = jj
                break
        tmp = vec[ii]
        vec[ii] = vec[idx1]
        vec[idx1] = tmp
        ipiv[ii] = idx1

    ipiv = ipiv-1 # C 0-based indexing
    return ipiv


def print_casadi_expression(f: Union[MX, SX, DM]):
    for ii in range(casadi_length(f)):
        print(f[ii,:])


def verbose_system_call(cmd, verbose=True, shell=False):
    return call(
        cmd,
        stdout=None if verbose else DEVNULL,
        stderr=None if verbose else STDOUT,
        shell=shell
    )



#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from setuptools import setup, find_packages

import sys
print(sys.version_info)

if sys.version_info < (3,5):
    sys.exit('Python version 3.5 or later required. Exiting.')

setup(name='acados_template',
    version='0.1',
    python_requires='>=3.5',
    description='A templating framework for acados',
    url='http://github.com/acados/acados',
    author='Andrea Zanelli',
    maintainer="Jonathan Frey",
    license='BSD 2-Clause',
    packages = find_packages(),
    include_package_data = True,
    setup_requires=['setuptools_scm'],
    use_scm_version={
      "fallback_version": "0.1-local",
      "root": "../..",
      "relative_to": __file__
    },
    install_requires=[
       'numpy',
       'scipy',
       'casadi',
       'matplotlib',
       'future-fstrings',
       'cython',
    ],
    package_data={'': [
        'simulink_default_opts.json',
        'c_templates_tera/CMakeLists.in.txt',
        'c_templates_tera/Makefile.in',
        'c_templates_tera/acados_sim_solver.in.c',
        'c_templates_tera/acados_sim_solver.in.h',
        'c_templates_tera/acados_sim_solver.in.pxd',
        'c_templates_tera/acados_solver.in.c',
        'c_templates_tera/acados_solver.in.h',
        'c_templates_tera/acados_solver.in.pxd',
        'c_templates_tera/constraints.in.h',
        'c_templates_tera/cost.in.h',
        'c_templates_tera/main.in.c',
        'c_templates_tera/main_sim.in.c',
        'c_templates_tera/model.in.h',
        'c_templates_tera/matlab_templates/acados_mex_create.in.c',
        'c_templates_tera/matlab_templates/acados_mex_free.in.c',
        'c_templates_tera/matlab_templates/acados_mex_set.in.c',
        'c_templates_tera/matlab_templates/acados_mex_solve.in.c',
        'c_templates_tera/matlab_templates/acados_sim_solver_sfun.in.c',
        'c_templates_tera/matlab_templates/acados_solver_sfun.in.c',
        'c_templates_tera/matlab_templates/make_mex.in.m',
        'c_templates_tera/matlab_templates/make_sfun.in.m',
        'c_templates_tera/matlab_templates/make_sfun_sim.in.m',
        'c_templates_tera/matlab_templates/mex_solver.in.m',
        'custom_update_templates/custom_update_function_zoro_template.in.c',
        'custom_update_templates/custom_update_function_zoro_template.in.h',
        ]},
    zip_safe=False
)



###################################################################################################
#                                                                                                 #
# This file is part of HPIPM.                                                                     #
#                                                                                                 #
# HPIPM -- High-Performance Interior Point Method.                                                #
# Copyright (C) 2019 by Gianluca Frison.                                                          #
# Developed at IMTEK (University of Freiburg) under the supervision of Moritz Diehl.              #
# All rights reserved.                                                                            #
#                                                                                                 #
# The 2-Clause BSD License                                                                        #
#                                                                                                 #
# Redistribution and use in source and binary forms, with or without                              #
# modification, are permitted provided that the following conditions are met:                     #
#                                                                                                 #
# 1. Redistributions of source code must retain the above copyright notice, this                  #
#    list of conditions and the following disclaimer.                                             #
# 2. Redistributions in binary form must reproduce the above copyright notice,                    #
#    this list of conditions and the following disclaimer in the documentation                    #
#    and/or other materials provided with the distribution.                                       #
#                                                                                                 #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND                 #
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED                   #
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE                          #
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR                 #
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES                  #
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;                    #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND                     #
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT                      #
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS                   #
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                    #
#                                                                                                 #
# Author: Gianluca Frison, gianluca.frison (at) imtek.uni-freiburg.de                             #
#                                                                                                 #
###################################################################################################

from ctypes import *
import ctypes.util 
import numpy as np


class hpipm_ocp_qp_dim:


	def __init__(self, N):

		# load hpipm shared library
		__hpipm   = CDLL('libhpipm.so')
		self.__hpipm = __hpipm

		# C dim struct
		dim_struct_size = __hpipm.d_ocp_qp_dim_strsize()
		dim_struct = cast(create_string_buffer(dim_struct_size), c_void_p)
		self.dim_struct = dim_struct

		# C dim internal memory
		dim_mem_size = __hpipm.d_ocp_qp_dim_memsize(N)
		dim_mem = cast(create_string_buffer(dim_mem_size), c_void_p)
		self.dim_mem = dim_mem

		# create C dim
		__hpipm.d_ocp_qp_dim_create(N, self.dim_struct, self.dim_mem)


	def set(self, field, value, idx_start, idx_end=None):
		self.__hpipm.d_ocp_qp_dim_set.argtypes = [c_char_p, c_int, c_int, c_void_p]
		field_name_b = field.encode('utf-8')
		if idx_end==None:
			self.__hpipm.d_ocp_qp_dim_set(c_char_p(field_name_b), idx_start, value, self.dim_struct)
		else:
			for i in range(idx_start, idx_end+1):
				self.__hpipm.d_ocp_qp_dim_set(c_char_p(field_name_b), i, value, self.dim_struct)
		return


	def print_C_struct(self):
		self.__hpipm.d_ocp_qp_dim_print(self.dim_struct)
		return 


	def codegen(self, file_name, mode):
		file_name_b = file_name.encode('utf-8')
		mode_b = mode.encode('utf-8')
		self.__hpipm.d_ocp_qp_dim_codegen(c_char_p(file_name_b), c_char_p(mode_b), self.dim_struct)
		return




from pylab import *
import unittest

from rockit import Ocp, DirectMethod, MultipleShooting, FreeTime, Stage, external_method
import numpy as np
from casadi import kron, DM
import casadi as ca
from rockit.casadi_helpers import AutoBrancher

class FatropTests(unittest.TestCase):
    def test_hello_world(self):

        ocp = Ocp(t0=0, T=10)

        x1 = ocp.state()
        x2 = ocp.state()

        u = ocp.control()


        e = 1 - x1**2

        ocp.set_der(x1, x2)
        ocp.set_der(x2,  e * x2 - x1 + u)

        ocp.add_objective(ocp.sum(x1**2 + x2**2 + u**2))
        ocp.add_objective(ocp.at_tf(x2**2))

        ocp.subject_to(x2 >= -0.25,include_last=False)
        ocp.subject_to(-1 <= (u <= 1 ))

        p = ocp.parameter()

        ocp.set_value(p, 1)
        # Boundary constraints
        ocp.subject_to(ocp.at_t0(x1) == p)
        ocp.subject_to(ocp.at_t0(x2) == 0)

        
        ocp.solver('ipopt', {"expand":True})
        ocp.method(MultipleShooting(N=20, intg = 'expl_euler'))

        sol = ocp.solve()
        _, x1_sampled_ref_p1 = sol.sample(x1, grid='control')
        _, x2_sampled_ref_p1 = sol.sample(x2, grid='control')
        _, u_sampled_ref_p1 = sol.sample(u, grid='control')

        ocp.set_value(p, 0.5)
        sol = ocp.solve()
        
        _, x1_sampled_ref_p05 = sol.sample(x1, grid='control')
        _, x2_sampled_ref_p05 = sol.sample(x2, grid='control')
        _, u_sampled_ref_p05 = sol.sample(u, grid='control')
        
        for ab in AutoBrancher():
            
            ocp = Ocp(t0=0, T=10)

            x1 = ocp.state()
            x2 = ocp.state()

            u = ocp.control()


            e = 1 - x1**2

            ocp.set_der(x1, x2)
            ocp.set_der(x2,  e * x2 - x1 + u)

            ocp.add_objective(ocp.sum(x1**2 + x2**2 + u**2))
            ocp.add_objective(ocp.at_tf(x2**2))

            ocp.subject_to(x2 >= -0.25,include_last=False)
            ocp.subject_to(-1 <= (u <= 1 ))

            p = ocp.parameter()

            ocp.set_value(p, 1)
            # Boundary constraints
            ocp.subject_to(ocp.at_t0(x1) == p)
            ocp.subject_to(ocp.at_t0(x2) == 0)

            
            method = external_method('fatrop',N=20, intg='expl_euler', mode = ab.branch(['fatropy','interface']))
            ocp.method(method)

            sol = ocp.solve()
            _, x1_sampled = sol.sample(x1, grid='control')
            _, x2_sampled = sol.sample(x2, grid='control')
            _, u_sampled = sol.sample(u, grid='control')
            np.testing.assert_allclose(x1_sampled, x1_sampled_ref_p1, atol=1e-6)
            np.testing.assert_allclose(x2_sampled, x2_sampled_ref_p1, atol=1e-6)
            np.testing.assert_allclose(u_sampled, u_sampled_ref_p1, atol=1e-6)
            

            ocp.set_value(p, 0.5)
            sol = ocp.solve()
            
            _, x1_sampled = sol.sample(x1, grid='control')
            _, x2_sampled = sol.sample(x2, grid='control')
            _, u_sampled = sol.sample(u, grid='control')
            np.testing.assert_allclose(x1_sampled, x1_sampled_ref_p05, atol=1e-6)
            np.testing.assert_allclose(x2_sampled, x2_sampled_ref_p05, atol=1e-6)
            np.testing.assert_allclose(u_sampled, u_sampled_ref_p05, atol=1e-6)
            
            _, x1_sampled = ocp.sample(x1, grid='control')
            _, x2_sampled = ocp.sample(x2, grid='control')
            _, u_sampled = ocp.sample(u, grid='control-')

            print([x1_sampled,x2_sampled,u_sampled,p])
            func = ocp.to_function("test", [x1_sampled,x2_sampled,u_sampled,p], [x1_sampled])
            
            x1_sampled = func(ca.DM.zeros(21,1), ca.DM.zeros(21,1), ca.DM.zeros(20,1),1)
            np.testing.assert_allclose(np.array(x1_sampled).squeeze(), x1_sampled_ref_p1, atol=1e-6)
            x1_sampled = func(ca.DM.zeros(21,1), ca.DM.zeros(21,1), ca.DM.zeros(20,1),0.5)
            np.testing.assert_allclose(np.array(x1_sampled).squeeze(), x1_sampled_ref_p05, atol=1e-6)

if __name__ == '__main__':
    unittest.main()




# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

import numpy as np
from .utils import J_to_idx, print_J_to_idx_note, J_to_idx_slack, check_if_nparray_and_flatten, check_if_2d_nparray

class AcadosOcpConstraints:
    """
    class containing the description of the constraints
    """
    def __init__(self):
        self.__constr_type_0 = 'BGH'
        self.__constr_type = 'BGH'
        self.__constr_type_e = 'BGH'
        # initial x
        self.__lbx_0   = np.array([])
        self.__ubx_0   = np.array([])
        self.__idxbx_0 = np.array([])
        self.__idxbxe_0 = np.array([])
        self.__has_x0 = False
        # state bounds
        self.__lbx     = np.array([])
        self.__ubx     = np.array([])
        self.__idxbx   = np.array([])
        # bounds on x at shooting node N
        self.__lbx_e   = np.array([])
        self.__ubx_e   = np.array([])
        self.__idxbx_e = np.array([])
        # bounds on u
        self.__lbu     = np.array([])
        self.__ubu     = np.array([])
        self.__idxbu   = np.array([])
        # polytopic constraints
        self.__lg      = np.array([])
        self.__ug      = np.array([])
        self.__D       = np.zeros((0,0))
        self.__C       = np.zeros((0,0))
        # polytopic constraints at shooting node N
        self.__lg_e    = np.array([])
        self.__ug_e    = np.array([])
        self.__C_e     = np.zeros((0,0))
        # nonlinear constraints at initial shooting node
        self.__lh_0    = np.array([])
        self.__uh_0    = np.array([])
        # nonlinear constraints
        self.__lh      = np.array([])
        self.__uh      = np.array([])
        # nonlinear constraints at shooting node N
        self.__lh_e    = np.array([])
        self.__uh_e    = np.array([])
        # convex-over-nonlinear constraints
        self.__lphi_0 = np.array([])
        self.__uphi_0 = np.array([])
        self.__lphi    = np.array([])
        self.__uphi    = np.array([])
        self.__lphi_e = np.array([])
        self.__uphi_e = np.array([])
        # SLACK BOUNDS
        # soft bounds on x
        self.__lsbx   = np.array([])
        self.__usbx   = np.array([])
        self.__idxsbx = np.array([])
        # soft bounds on u
        self.__lsbu   = np.array([])
        self.__usbu   = np.array([])
        self.__idxsbu = np.array([])
        # soft bounds on x at shooting node N
        self.__lsbx_e  = np.array([])
        self.__usbx_e  = np.array([])
        self.__idxsbx_e= np.array([])
        # soft bounds on general linear constraints
        self.__lsg    = np.array([])
        self.__usg    = np.array([])
        self.__idxsg  = np.array([])
        # soft bounds on general linear constraints at shooting node N
        self.__lsg_e    = np.array([])
        self.__usg_e    = np.array([])
        self.__idxsg_e  = np.array([])
        # soft bounds on nonlinear constraints at shooting node 0
        self.__lsh_0    = np.array([])
        self.__ush_0    = np.array([])
        self.__idxsh_0  = np.array([])
        # soft bounds on nonlinear constraints
        self.__lsh    = np.array([])
        self.__ush    = np.array([])
        self.__idxsh  = np.array([])
        # soft bounds on nonlinear constraints at shooting node N
        self.__lsh_e    = np.array([])
        self.__ush_e    = np.array([])
        self.__idxsh_e  = np.array([])
        # soft bounds on convex-over-nonlinear constraint (BGP) at shooting node 0
        self.__lsphi_0    = np.array([])
        self.__usphi_0    = np.array([])
        self.__idxsphi_0  = np.array([])
        # soft bounds on convex-over-nonlinear constraint (BGP)
        self.__lsphi  = np.array([])
        self.__usphi  = np.array([])
        self.__idxsphi  = np.array([])
        # soft bounds on convex-over-nonlinear constraint (BGP) at shooting node N
        self.__lsphi_e    = np.array([])
        self.__usphi_e    = np.array([])
        self.__idxsphi_e  = np.array([])


    # types
    @property
    def constr_type(self):
        """Constraints type for shooting nodes (0 to N-1). string in {BGH, BGP}.
        Detected automatically from model.
        Default: BGH; BGP is for convex over nonlinear.
        """
        return self.__constr_type

    @property
    def constr_type_0(self):
        """Constraints type for initial shooting node. string in {BGH, BGP}.
        Detected automatically from model.
        Default: BGH; BGP is for convex over nonlinear.
        """
        return self.__constr_type_0

    @property
    def constr_type_e(self):
        """Constraints type for terminal shooting node N. string in {BGH, BGP}.
        Detected automatically from model.
        Default: BGH; BGP is for convex over nonlinear.
        """
        return self.__constr_type_e

    # initial bounds on x
    @property
    def lbx_0(self):
        r""":math:`\underline{x_0}` - lower bounds on x at initial stage 0.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`."""
        return self.__lbx_0

    @property
    def ubx_0(self):
        r""":math:`\bar{x_0}` - upper bounds on x at initial stage 0.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__ubx_0

    @property
    def Jbx_0(self):
        """:math:`J_{bx,0}` - matrix coefficient for bounds on x at initial stage 0.
        Translated internally to :py:attr:`idxbx_0`"""
        print_J_to_idx_note()
        return self.__idxbx_0

    @property
    def idxbx_0(self):
        """Indices of bounds on x at initial stage 0
        -- can be set automatically via x0.
        Can be set by using :py:attr:`Jbx_0`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxbx_0

    @property
    def idxbxe_0(self):
        """Indices of bounds on x0 that are equalities -- can be set automatically via :py:attr:`x0`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxbxe_0

    # bounds on x
    @property
    def lbx(self):
        r""":math:`\underline{x}` - lower bounds on x at intermediate shooting nodes (1 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__lbx

    @property
    def ubx(self):
        r""":math:`\bar{x}` - upper bounds on x at intermediate shooting nodes (1 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__ubx

    @property
    def idxbx(self):
        """indices of bounds on x (defines :math:`J_{bx}`) at intermediate shooting nodes (1 to N-1).
        Can be set by using :py:attr:`Jbx`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxbx

    @property
    def Jbx(self):
        """:math:`J_{bx}` - matrix coefficient for bounds on x
        at intermediate shooting nodes (1 to N-1).
        Translated internally into :py:attr:`idxbx`."""
        print_J_to_idx_note()
        return self.__idxbx

    # bounds on x at shooting node N
    @property
    def lbx_e(self):
        r""":math:`\underline{x}^e` - lower bounds on x at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__lbx_e

    @property
    def ubx_e(self):
        r""":math:`\bar{x}^e` - upper bounds on x at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__ubx_e

    @property
    def idxbx_e(self):
        """Indices for bounds on x at terminal shooting node N (defines :math:`J_{bx}^e`).
        Can be set by using :py:attr:`Jbx_e`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxbx_e

    @property
    def Jbx_e(self):
        """:math:`J_{bx}^e` matrix coefficient for bounds on x at terminal shooting node N.
        Translated internally into :py:attr:`idxbx_e`."""
        print_J_to_idx_note()
        return self.__idxbx_e

    # bounds on u
    @property
    def lbu(self):
        r""":math:`\underline{u}` - lower bounds on u at shooting nodes (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`
        """
        return self.__lbu

    @property
    def ubu(self):
        r""":math:`\bar{u}` - upper bounds on u at shooting nodes (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`
        """
        return self.__ubu

    @property
    def idxbu(self):
        """Indices of bounds on u (defines :math:`J_{bu}`) at shooting nodes (0 to N-1).
        Can be set by using :py:attr:`Jbu`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`
        """
        return self.__idxbu

    @property
    def Jbu(self):
        """:math:`J_{bu}` - matrix coefficient for bounds on u at shooting nodes (0 to N-1).
        Translated internally to :py:attr:`idxbu`.
        """
        print_J_to_idx_note()
        return self.__idxbu

    # polytopic constraints
    @property
    def C(self):
        r""":math:`C` - C matrix in :math:`\underline{g} \leq D \, u + C \, x \leq \bar{g}`
        at shooting nodes (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array((0,0))`.
        """
        return self.__C

    @property
    def D(self):
        r""":math:`D` - D matrix in :math:`\underline{g} \leq D \, u + C \, x \leq \bar{g}`
        at shooting nodes (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array((0,0))`
        """
        return self.__D

    @property
    def lg(self):
        r""":math:`\underline{g}` - lower bound for general polytopic inequalities
        at shooting nodes (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`
        """
        return self.__lg

    @property
    def ug(self):
        r""":math:`\bar{g}` - upper bound for general polytopic inequalities
        at shooting nodes (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__ug

    # polytopic constraints at shooting node N
    @property
    def C_e(self):
        """:math:`C^e` - C matrix at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array((0,0))`.
        """
        return self.__C_e

    @property
    def lg_e(self):
        r""":math:`\underline{g}^e` - lower bound on general polytopic inequalities
        at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__lg_e

    @property
    def ug_e(self):
        r""":math:`\bar{g}^e` - upper bound on general polytopic inequalities
        at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__ug_e


    # nonlinear constraints
    @property
    def lh(self):
        r""":math:`\underline{h}` - lower bound for nonlinear inequalities
        at intermediate shooting nodes (1 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__lh

    @property
    def uh(self):
        r""":math:`\bar{h}` - upper bound for nonlinear inequalities
        at intermediate shooting nodes (1 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__uh

    # nonlinear constraints at initial shooting node
    @property
    def lh_0(self):
        r""":math:`\underline{h}^0` - lower bound on nonlinear inequalities
        at initial shooting node (0).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__lh_0

    @property
    def uh_0(self):
        r""":math:`\bar{h}^0` - upper bound on nonlinear inequalities
        at initial shooting node (0).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__uh_0

    # nonlinear constraints at shooting node N
    @property
    def lh_e(self):
        r""":math:`\underline{h}^e` - lower bound on nonlinear inequalities
        at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__lh_e

    @property
    def uh_e(self):
        r""":math:`\bar{h}^e` - upper bound on nonlinear inequalities
        at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__uh_e

    # convex-over-nonlinear constraints
    @property
    def lphi(self):
        r""":math:`\underline{\phi}` - lower bound for convex-over-nonlinear inequalities
        at shooting nodes (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__lphi

    @property
    def uphi(self):
        r""":math:`\bar{\phi}` - upper bound for convex-over-nonlinear inequalities
        at shooting nodes (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__uphi

    # convex-over-nonlinear constraints at shooting node N
    @property
    def lphi_e(self):
        r""":math:`\underline{\phi}^e` - lower bound on convex-over-nonlinear inequalities
        at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__lphi_e

    @property
    def uphi_e(self):
        r""":math:`\bar{\phi}^e` - upper bound on convex-over-nonlinear inequalities
        at terminal shooting node N.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__uphi_e

    @property
    def lphi_0(self):
        r""":math:`\underline{\phi}^0` - lower bound on convex-over-nonlinear inequalities
        at shooting node 0.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__lphi_0

    @property
    def uphi_0(self):
        r""":math:`\bar{\phi}^0` - upper bound on convex-over-nonlinear inequalities
        at shooting node 0.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`.
        """
        return self.__uphi_0


    # SLACK bounds
    # soft bounds on x
    @property
    def lsbx(self):
        """Lower bounds on slacks corresponding to soft lower bounds on x
        at stages (1 to N-1);
        not required - zeros by default"""
        return self.__lsbx

    @property
    def usbx(self):
        """Lower bounds on slacks corresponding to soft upper bounds on x
        at stages (1 to N-1);
        not required - zeros by default"""
        return self.__usbx

    @property
    def idxsbx(self):
        """Indices of soft bounds on x within the indices of bounds on x
        at stages (1 to N-1).
        Can be set by using :py:attr:`Jsbx`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxsbx

    @property
    def Jsbx(self):
        """:math:`J_{sbx}` - matrix coefficient for soft bounds on x
        at stages (1 to N-1);
        Translated internally into :py:attr:`idxsbx`."""
        print_J_to_idx_note()
        return self.__idxsbx

    # soft bounds on u
    @property
    def lsbu(self):
        """Lower bounds on slacks corresponding to soft lower bounds on u
        at stages (0 to N-1).
        Not required - zeros by default."""
        return self.__lsbu

    @property
    def usbu(self):
        """Lower bounds on slacks corresponding to soft upper bounds on u
        at stages (0 to N-1);
        not required - zeros by default"""
        return self.__usbu

    @property
    def idxsbu(self):
        """Indices of soft bounds on u within the indices of bounds on u
        at stages (0 to N-1).
        Can be set by using :py:attr:`Jsbu`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxsbu

    @property
    def Jsbu(self):
        """:math:`J_{sbu}` - matrix coefficient for soft bounds on u
        at stages (0 to N-1);
        internally translated into :py:attr:`idxsbu`"""
        print_J_to_idx_note()
        return self.__idxsbu

    # soft bounds on x at shooting node N
    @property
    def lsbx_e(self):
        """Lower bounds on slacks corresponding to soft lower bounds on x at shooting node N.
        Not required - zeros by default"""
        return self.__lsbx_e

    @property
    def usbx_e(self):
        """Lower bounds on slacks corresponding to soft upper bounds on x at shooting node N.
        Not required - zeros by default"""
        return self.__usbx_e

    @property
    def idxsbx_e(self):
        """Indices of soft bounds on x at shooting node N, within the indices of bounds on x at terminal shooting node N.
        Can be set by using :py:attr:`Jsbx_e`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxsbx_e

    @property
    def Jsbx_e(self):
        """:math:`J_{sbx}^e` - matrix coefficient for soft bounds on x at terminal shooting node N.
        Translated internally to :py:attr:`idxsbx_e`"""
        print_J_to_idx_note()
        return self.__idxsbx_e

    # soft general linear constraints
    @property
    def lsg(self):
        """Lower bounds on slacks corresponding to soft lower bounds for general linear constraints
        at stages (0 to N-1).
        Type: :code:`np.ndarray`; default: :code:`np.array([])`
        """
        return self.__lsg

    @property
    def usg(self):
        """Lower bounds on slacks corresponding to soft upper bounds for general linear constraints.
        Not required - zeros by default"""
        return self.__usg

    @property
    def idxsg(self):
        """Indices of soft general linear constraints within the indices of general linear constraints.
        Can be set by using :py:attr:`Jsg`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxsg

    @property
    def Jsg(self):
        """:math:`J_{sg}` - matrix coefficient for soft bounds on general linear constraints.
        Translated internally to :py:attr:`idxsg`"""
        print_J_to_idx_note()
        return self.__idxsg

    # soft nonlinear constraints
    @property
    def lsh(self):
        """Lower bounds on slacks corresponding to soft lower bounds for nonlinear constraints.
        Not required - zeros by default"""
        return self.__lsh

    @property
    def ush(self):
        """Lower bounds on slacks corresponding to soft upper bounds for nonlinear constraints.
        Not required - zeros by default"""
        return self.__ush

    @property
    def idxsh(self):
        """Indices of soft nonlinear constraints within the indices of nonlinear constraints.
        Can be set by using :py:attr:`Jbx`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxsh

    @property
    def Jsh(self):
        """:math:`J_{sh}` - matrix coefficient for soft bounds on nonlinear constraints.
        Translated internally to :py:attr:`idxsh`"""
        print_J_to_idx_note()
        return self.__idxsh

    # soft bounds on convex-over-nonlinear constraints
    @property
    def lsphi(self):
        """Lower bounds on slacks corresponding to soft lower bounds for convex-over-nonlinear constraints.
        Not required - zeros by default"""
        return self.__lsphi

    @property
    def usphi(self):
        """Lower bounds on slacks corresponding to soft upper bounds for convex-over-nonlinear constraints.
        Not required - zeros by default"""
        return self.__usphi

    @property
    def idxsphi(self):
        """Indices of soft convex-over-nonlinear constraints within the indices of nonlinear constraints.
        Can be set by using :py:attr:`Jsphi`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxsphi

    @property
    def Jsphi(self):
        r""":math:`J_{s, \phi}` - matrix coefficient for soft bounds on convex-over-nonlinear constraints.
        Translated internally into :py:attr:`idxsphi`."""
        print_J_to_idx_note()
        return self.__idxsphi


    # soft bounds on general linear constraints at shooting node N
    @property
    def lsg_e(self):
        """Lower bounds on slacks corresponding to soft lower bounds for general linear constraints at shooting node N.
        Not required - zeros by default"""
        return self.__lsg_e

    @property
    def usg_e(self):
        """Lower bounds on slacks corresponding to soft upper bounds for general linear constraints at shooting node N.
        Not required - zeros by default"""
        return self.__usg_e

    @property
    def idxsg_e(self):
        """Indices of soft general linear constraints at shooting node N within the indices of general linear constraints at shooting node N.
        Can be set by using :py:attr:`Jsg_e`."""
        return self.__idxsg_e

    @property
    def Jsg_e(self):
        """:math:`J_{s,h}^e` - matrix coefficient for soft bounds on general linear constraints at terminal shooting node N.
        Translated internally to :py:attr:`idxsg_e`"""
        print_J_to_idx_note()
        return self.__idxsg_e


    # soft bounds on nonlinear constraints at shooting node 0
    @property
    def lsh_0(self):
        """Lower bounds on slacks corresponding to soft lower bounds for nonlinear constraints at initial shooting node 0.
        Not required - zeros by default"""
        return self.__lsh_0

    @property
    def ush_0(self):
        """Lower bounds on slacks corresponding to soft upper bounds for nonlinear constraints at initial shooting node 0.
        Not required - zeros by default"""
        return self.__ush_0

    @property
    def idxsh_0(self):
        """Indices of soft nonlinear constraints at shooting node N within the indices of nonlinear constraints at initial shooting node 0.
        Can be set by using :py:attr:`Jsh_0`."""
        return self.__idxsh_0

    @property
    def Jsh_0(self):
        """:math:`J_{s,h}^0` - matrix coefficient for soft bounds on nonlinear constraints at initial shooting node 0; fills :py:attr:`idxsh_0`"""
        print_J_to_idx_note()
        return self.__idxsh_0

    # soft bounds on convex-over-nonlinear constraints at shooting node N
    @property
    def lsphi_0(self):
        """Lower bounds on slacks corresponding to soft lower bounds for convex-over-nonlinear constraints at initial shooting node 0.
        Not required - zeros by default"""
        return self.__lsphi_0

    @property
    def usphi_0(self):
        """Lower bounds on slacks corresponding to soft upper bounds for convex-over-nonlinear constraints at initial shooting node 0.
        Not required - zeros by default"""
        return self.__usphi_0

    @property
    def idxsphi_0(self):
        """Indices of soft nonlinear constraints at shooting node N within the indices of nonlinear constraints at initial shooting node 0.
        Can be set by using :py:attr:`Jsphi_0`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxsphi_0

    @property
    def Jsphi_0(self):
        """:math:`J_{sh}^0` - matrix coefficient for soft bounds on convex-over-nonlinear constraints at shooting node N.
        Translated internally to :py:attr:`idxsphi_0`"""
        print_J_to_idx_note()
        return self.__idxsphi_0


    # soft bounds on nonlinear constraints at shooting node N
    @property
    def lsh_e(self):
        """Lower bounds on slacks corresponding to soft lower bounds for nonlinear constraints at terminal shooting node N.
        Not required - zeros by default"""
        return self.__lsh_e

    @property
    def ush_e(self):
        """Lower bounds on slacks corresponding to soft upper bounds for nonlinear constraints at terminal shooting node N.
        Not required - zeros by default"""
        return self.__ush_e

    @property
    def idxsh_e(self):
        """Indices of soft nonlinear constraints at shooting node N within the indices of nonlinear constraints at terminal shooting node N.
        Can be set by using :py:attr:`Jsh_e`."""
        return self.__idxsh_e

    @property
    def Jsh_e(self):
        """:math:`J_{s,h}^e` - matrix coefficient for soft bounds on nonlinear constraints at terminal shooting node N; fills :py:attr:`idxsh_e`"""
        print_J_to_idx_note()
        return self.__idxsh_e

    # soft bounds on convex-over-nonlinear constraints at shooting node N
    @property
    def lsphi_e(self):
        """Lower bounds on slacks corresponding to soft lower bounds for convex-over-nonlinear constraints at terminal shooting node N.
        Not required - zeros by default"""
        return self.__lsphi_e

    @property
    def usphi_e(self):
        """Lower bounds on slacks corresponding to soft upper bounds for convex-over-nonlinear constraints at terminal shooting node N.
        Not required - zeros by default"""
        return self.__usphi_e

    @property
    def idxsphi_e(self):
        """Indices of soft nonlinear constraints at shooting node N within the indices of nonlinear constraints at terminal shooting node N.
        Can be set by using :py:attr:`Jsphi_e`.
        Type: :code:`np.ndarray`; default: :code:`np.array([])`"""
        return self.__idxsphi_e

    @property
    def Jsphi_e(self):
        """:math:`J_{sh}^e` - matrix coefficient for soft bounds on convex-over-nonlinear constraints at shooting node N.
        Translated internally to :py:attr:`idxsphi_e`"""
        print_J_to_idx_note()
        return self.__idxsphi_e

    @property
    def x0(self):
        r"""
        :math:`x_0 \in \mathbb{R}^{n_x}` - initial state --
        Translated internally to :py:attr:`idxbx_0`, :py:attr:`lbx_0`, :py:attr:`ubx_0`, :py:attr:`idxbxe_0`
        """
        if self.has_x0:
            return self.lbx_0
        else:
            print("x0 is not set. You can set it or specify lbx_0, ubx_0, idxbx_0, idxbxe_0 to implement general bounds on x0.")
            print("")
            print("idxbx_0: ", self.__idxbx_0)
            print("lbx_0: ", self.__lbx_0)
            print("ubx_0: ", self.__ubx_0)
            print("idxbxe_0: ", self.__idxbxe_0)
            return None

    @property
    def has_x0(self):
        """
        Internal variable to check if x0 is set.
        Cannot be set from outside.
        :bool: True if x0 is set, False otherwise.
        """
        return self.__has_x0

    # SETTERS
    @constr_type.setter
    def constr_type(self, constr_type):
        constr_types = ('BGH', 'BGP')
        if constr_type in constr_types:
            self.__constr_type = constr_type
        else:
            raise Exception('Invalid constr_type value. Possible values are:\n\n' \
                    + ',\n'.join(constr_types) + '.\n\nYou have: ' + constr_type + '.\n\n')

    @constr_type_0.setter
    def constr_type_0(self, constr_type_0):
        constr_types = ('BGH', 'BGP')
        if constr_type_0 in constr_types:
            self.__constr_type_0 = constr_type_0
        else:
            raise Exception('Invalid constr_type_0 value. Possible values are:\n\n' \
                    + ',\n'.join(constr_types) + '.\n\nYou have: ' + constr_type_0 + '.\n\n')

    @constr_type_e.setter
    def constr_type_e(self, constr_type_e):
        constr_types = ('BGH', 'BGP')
        if constr_type_e in constr_types:
            self.__constr_type_e = constr_type_e
        else:
            raise Exception('Invalid constr_type_e value. Possible values are:\n\n' \
                    + ',\n'.join(constr_types) + '.\n\nYou have: ' + constr_type_e + '.\n\n')

    # initial x
    @lbx_0.setter
    def lbx_0(self, lbx_0):
        lbx_0 = check_if_nparray_and_flatten(lbx_0, "lbx_0")
        self.__lbx_0 = lbx_0

    @ubx_0.setter
    def ubx_0(self, ubx_0):
        ubx_0 = check_if_nparray_and_flatten(ubx_0, "ubx_0")
        self.__ubx_0 = ubx_0

    @idxbx_0.setter
    def idxbx_0(self, idxbx_0):
        idxbx_0 = check_if_nparray_and_flatten(idxbx_0, "idxbx_0")
        self.__idxbx_0 = idxbx_0

    @Jbx_0.setter
    def Jbx_0(self, Jbx_0):
        if isinstance(Jbx_0, np.ndarray):
            self.__idxbx_0 = J_to_idx(Jbx_0)
        else:
            raise Exception('Invalid Jbx_0 value.')

    @idxbxe_0.setter
    def idxbxe_0(self, idxbxe_0):
        idxbxe_0 = check_if_nparray_and_flatten(idxbxe_0, "idxbxe_0")
        self.__idxbxe_0 = idxbxe_0

    @x0.setter
    def x0(self, x0):
        x0 = check_if_nparray_and_flatten(x0, "x0")
        self.__lbx_0 = x0
        self.__ubx_0 = x0
        self.__idxbx_0 = np.arange(x0.size)
        self.__idxbxe_0 = np.arange(x0.size)
        self.__has_x0 = True

    # bounds on x
    @lbx.setter
    def lbx(self, lbx):
        lbx = check_if_nparray_and_flatten(lbx, "lbx")
        self.__lbx = lbx

    @ubx.setter
    def ubx(self, ubx):
        ubx = check_if_nparray_and_flatten(ubx, "ubx")
        self.__ubx = ubx

    @idxbx.setter
    def idxbx(self, idxbx):
        idxbx = check_if_nparray_and_flatten(idxbx, "idxbx")
        self.__idxbx = idxbx

    @Jbx.setter
    def Jbx(self, Jbx):
        if isinstance(Jbx, np.ndarray):
            self.__idxbx = J_to_idx(Jbx)
        else:
            raise Exception('Invalid Jbx value.')

    # bounds on u
    @lbu.setter
    def lbu(self, lbu):
        lbu = check_if_nparray_and_flatten(lbu, "lbu")
        self.__lbu = lbu

    @ubu.setter
    def ubu(self, ubu):
        ubu = check_if_nparray_and_flatten(ubu, "ubu")
        self.__ubu = ubu

    @idxbu.setter
    def idxbu(self, idxbu):
        idxbu = check_if_nparray_and_flatten(idxbu, "idxbu")
        self.__idxbu = idxbu

    @Jbu.setter
    def Jbu(self, Jbu):
        if isinstance(Jbu, np.ndarray):
            self.__idxbu = J_to_idx(Jbu)
        else:
            raise Exception('Invalid Jbu value.')

    # bounds on x at shooting node N
    @lbx_e.setter
    def lbx_e(self, lbx_e):
        lbx_e = check_if_nparray_and_flatten(lbx_e, "lbx_e")
        self.__lbx_e = lbx_e

    @ubx_e.setter
    def ubx_e(self, ubx_e):
        ubx_e = check_if_nparray_and_flatten(ubx_e, "ubx_e")
        self.__ubx_e = ubx_e

    @idxbx_e.setter
    def idxbx_e(self, idxbx_e):
        idxbx_e = check_if_nparray_and_flatten(idxbx_e, "idxbx_e")
        self.__idxbx_e = idxbx_e

    @Jbx_e.setter
    def Jbx_e(self, Jbx_e):
        if isinstance(Jbx_e, np.ndarray):
            self.__idxbx_e = J_to_idx(Jbx_e)
        else:
            raise Exception('Invalid Jbx_e value.')

    # polytopic constraints
    @D.setter
    def D(self, D):
        check_if_2d_nparray(D, "D")
        self.__D = D

    @C.setter
    def C(self, C):
        check_if_2d_nparray(C, "C")
        self.__C = C

    # polytopic constraints at shooting node N
    @C_e.setter
    def C_e(self, C_e):
        check_if_2d_nparray(C_e, "C_e")
        self.__C_e = C_e

    @lg.setter
    def lg(self, value):
        value = check_if_nparray_and_flatten(value, 'lg')
        self.__lg = value

    @ug.setter
    def ug(self, value):
        value = check_if_nparray_and_flatten(value, 'ug')
        self.__ug = value

    @lg_e.setter
    def lg_e(self, value):
        value = check_if_nparray_and_flatten(value, 'lg_e')
        self.__lg_e = value

    @ug_e.setter
    def ug_e(self, value):
        value = check_if_nparray_and_flatten(value, 'ug_e')
        self.__ug_e = value

    # nonlinear constraints
    @lh.setter
    def lh(self, value):
        value = check_if_nparray_and_flatten(value, 'lh')
        self.__lh = value

    @uh.setter
    def uh(self, value):
        value = check_if_nparray_and_flatten(value, 'uh')
        self.__uh = value

    @lh_e.setter
    def lh_e(self, value):
        value = check_if_nparray_and_flatten(value, 'lh_e')
        self.__lh_e = value

    @uh_e.setter
    def uh_e(self, value):
        value = check_if_nparray_and_flatten(value, 'uh_e')
        self.__uh_e = value

    @lh_0.setter
    def lh_0(self, value):
        value = check_if_nparray_and_flatten(value, 'lh_0')
        self.__lh_0 = value

    @uh_0.setter
    def uh_0(self, value):
        value = check_if_nparray_and_flatten(value, 'uh_0')
        self.__uh_0 = value

    # convex-over-nonlinear constraints
    @lphi.setter
    def lphi(self, value):
        value = check_if_nparray_and_flatten(value, 'lphi')
        self.__lphi = value

    @uphi.setter
    def uphi(self, value):
        value = check_if_nparray_and_flatten(value, 'uphi')
        self.__uphi = value

    @lphi_e.setter
    def lphi_e(self, value):
        value = check_if_nparray_and_flatten(value, 'lphi_e')
        self.__lphi_e = value

    @uphi_e.setter
    def uphi_e(self, value):
        value = check_if_nparray_and_flatten(value, 'uphi_e')
        self.__uphi_e = value

    @lphi_0.setter
    def lphi_0(self, value):
        value = check_if_nparray_and_flatten(value, 'lphi_0')
        self.__lphi_0 = value

    @uphi_0.setter
    def uphi_0(self, value):
        value = check_if_nparray_and_flatten(value, 'uphi_0')
        self.__uphi_0 = value

    # SLACK bounds
    # soft bounds on x
    @lsbx.setter
    def lsbx(self, value):
        value = check_if_nparray_and_flatten(value, 'lsbx')
        self.__lsbx = value

    @usbx.setter
    def usbx(self, value):
        value = check_if_nparray_and_flatten(value, 'usbx')
        self.__usbx = value

    @idxsbx.setter
    def idxsbx(self, idxsbx):
        idxsbx = check_if_nparray_and_flatten(idxsbx, "idxsbx")
        self.__idxsbx = idxsbx

    @Jsbx.setter
    def Jsbx(self, Jsbx):
        if isinstance(Jsbx, np.ndarray):
            self.__idxsbx = J_to_idx_slack(Jsbx)
        else:
            raise Exception('Invalid Jsbx value, expected numpy array.')

    # soft bounds on u
    @lsbu.setter
    def lsbu(self, value):
        value = check_if_nparray_and_flatten(value, 'lsbu')
        self.__lsbu = value

    @usbu.setter
    def usbu(self, value):
        value = check_if_nparray_and_flatten(value, 'usbu')
        self.__usbu = value

    @idxsbu.setter
    def idxsbu(self, idxsbu):
        idxsbu = check_if_nparray_and_flatten(idxsbu, "idxsbu")
        self.__idxsbu = idxsbu

    @Jsbu.setter
    def Jsbu(self, Jsbu):
        if isinstance(Jsbu, np.ndarray):
            self.__idxsbu = J_to_idx_slack(Jsbu)
        else:
            raise Exception('Invalid Jsbu value.')

    # soft bounds on x at shooting node N
    @lsbx_e.setter
    def lsbx_e(self, value):
        value = check_if_nparray_and_flatten(value, 'lsbx_e')
        self.__lsbx_e = value

    @usbx_e.setter
    def usbx_e(self, value):
        value = check_if_nparray_and_flatten(value, 'usbx_e')
        self.__usbx_e = value

    @idxsbx_e.setter
    def idxsbx_e(self, idxsbx_e):
        idxsbx_e = check_if_nparray_and_flatten(idxsbx_e, "idxsbx_e")
        self.__idxsbx_e = idxsbx_e

    @Jsbx_e.setter
    def Jsbx_e(self, Jsbx_e):
        if isinstance(Jsbx_e, np.ndarray):
            self.__idxsbx_e = J_to_idx_slack(Jsbx_e)
        else:
            raise Exception('Invalid Jsbx_e value.')

    # soft bounds on general linear constraints
    @lsg.setter
    def lsg(self, value):
        value = check_if_nparray_and_flatten(value, 'lsg')
        self.__lsg = value

    @usg.setter
    def usg(self, value):
        value = check_if_nparray_and_flatten(value, 'usg')
        self.__usg = value

    @idxsg.setter
    def idxsg(self, value):
        value = check_if_nparray_and_flatten(value, 'idxsg')
        self.__idxsg = value

    @Jsg.setter
    def Jsg(self, Jsg):
        if isinstance(Jsg, np.ndarray):
            self.__idxsg = J_to_idx_slack(Jsg)
        else:
            raise Exception('Invalid Jsg value, expected numpy array.')

    # soft bounds on nonlinear constraints
    @lsh.setter
    def lsh(self, value):
        value = check_if_nparray_and_flatten(value, 'lsh')
        self.__lsh = value

    @ush.setter
    def ush(self, value):
        value = check_if_nparray_and_flatten(value, 'ush')
        self.__ush = value

    @idxsh.setter
    def idxsh(self, value):
        value = check_if_nparray_and_flatten(value, 'idxsh')
        self.__idxsh = value


    @Jsh.setter
    def Jsh(self, Jsh):
        if isinstance(Jsh, np.ndarray):
            self.__idxsh = J_to_idx_slack(Jsh)
        else:
            raise Exception('Invalid Jsh value, expected numpy array.')

    # soft bounds on convex-over-nonlinear constraints
    @lsphi.setter
    def lsphi(self, value):
        value = check_if_nparray_and_flatten(value, 'lsphi')
        self.__lsphi = value

    @usphi.setter
    def usphi(self, value):
        value = check_if_nparray_and_flatten(value, 'usphi')
        self.__usphi = value

    @idxsphi.setter
    def idxsphi(self, value):
        value = check_if_nparray_and_flatten(value, 'idxsphi')
        self.__idxsphi = value

    @Jsphi.setter
    def Jsphi(self, Jsphi):
        if isinstance(Jsphi, np.ndarray):
            self.__idxsphi = J_to_idx_slack(Jsphi)
        else:
            raise Exception('Invalid Jsphi value, expected numpy array.')

    # soft bounds on general linear constraints at shooting node N
    @lsg_e.setter
    def lsg_e(self, value):
        value = check_if_nparray_and_flatten(value, 'lsg_e')
        self.__lsg_e = value

    @usg_e.setter
    def usg_e(self, value):
        value = check_if_nparray_and_flatten(value, 'usg_e')
        self.__usg_e = value

    @idxsg_e.setter
    def idxsg_e(self, value):
        value = check_if_nparray_and_flatten(value, 'idxsg_e')
        self.__idxsg_e = value

    @Jsg_e.setter
    def Jsg_e(self, Jsg_e):
        if isinstance(Jsg_e, np.ndarray):
            self.__idxsg_e = J_to_idx_slack(Jsg_e)
        else:
            raise Exception('Invalid Jsg_e value, expected numpy array.')

    # soft bounds on nonlinear constraints at shooting node N
    @lsh_e.setter
    def lsh_e(self, value):
        value = check_if_nparray_and_flatten(value, 'lsh_e')
        self.__lsh_e = value

    @ush_e.setter
    def ush_e(self, value):
        value = check_if_nparray_and_flatten(value, 'ush_e')
        self.__ush_e = value

    @idxsh_e.setter
    def idxsh_e(self, value):
        value = check_if_nparray_and_flatten(value, 'idxsh_e')
        self.__idxsh_e = value

    @Jsh_e.setter
    def Jsh_e(self, Jsh_e):
        if isinstance(Jsh_e, np.ndarray):
            self.__idxsh_e = J_to_idx_slack(Jsh_e)
        else:
            raise Exception('Invalid Jsh_e value, expected numpy array.')


    # soft bounds on convex-over-nonlinear constraints at shooting node N
    @lsphi_e.setter
    def lsphi_e(self, value):
        value = check_if_nparray_and_flatten(value, 'lsphi_e')
        self.__lsphi_e = value

    @usphi_e.setter
    def usphi_e(self, value):
        value = check_if_nparray_and_flatten(value, 'usphi_e')
        self.__usphi_e = value

    @idxsphi_e.setter
    def idxsphi_e(self, value):
        value = check_if_nparray_and_flatten(value, 'idxsphi_e')
        self.__idxsphi_e = value

    @Jsphi_e.setter
    def Jsphi_e(self, Jsphi_e):
        if isinstance(Jsphi_e, np.ndarray):
            self.__idxsphi_e = J_to_idx_slack(Jsphi_e)
        else:
            raise Exception('Invalid Jsphi_e value.')

    # soft constraints at shooting node 0
    @lsh_0.setter
    def lsh_0(self, value):
        value = check_if_nparray_and_flatten(value, 'lsh_0')
        self.__lsh_0 = value

    @ush_0.setter
    def ush_0(self, value):
        value = check_if_nparray_and_flatten(value, 'ush_0')
        self.__ush_0 = value

    @idxsh_0.setter
    def idxsh_0(self, value):
        value = check_if_nparray_and_flatten(value, 'idxsh_0')
        self.__idxsh_0 = value

    @Jsh_0.setter
    def Jsh_0(self, Jsh_0):
        if isinstance(Jsh_0, np.ndarray):
            self.__idxsh_0 = J_to_idx_slack(Jsh_0)
        else:
            raise Exception('Invalid Jsh_0 value, expected numpy array.')

    @lsphi_0.setter
    def lsphi_0(self, value):
        value = check_if_nparray_and_flatten(value, 'lsphi_0')
        self.__lsphi_0 = value

    @usphi_0.setter
    def usphi_0(self, value):
        value = check_if_nparray_and_flatten(value, 'usphi_0')
        self.__usphi_0 = value

    @idxsphi_0.setter
    def idxsphi_0(self, value):
        value = check_if_nparray_and_flatten(value, 'idxsphi_0')
        self.__idxsphi_0 = value

    @Jsphi_0.setter
    def Jsphi_0(self, Jsphi_0):
        if isinstance(Jsphi_0, np.ndarray):
            self.__idxsphi_0 = J_to_idx_slack(Jsphi_0)
        else:
            raise Exception('Invalid Jsphi_0 value.')

    def set(self, attr, value):
        setattr(self, attr, value)



###################################################################################################
#                                                                                                 #
# This file is part of HPIPM.                                                                     #
#                                                                                                 #
# HPIPM -- High-Performance Interior Point Method.                                                #
# Copyright (C) 2019 by Gianluca Frison.                                                          #
# Developed at IMTEK (University of Freiburg) under the supervision of Moritz Diehl.              #
# All rights reserved.                                                                            #
#                                                                                                 #
# The 2-Clause BSD License                                                                        #
#                                                                                                 #
# Redistribution and use in source and binary forms, with or without                              #
# modification, are permitted provided that the following conditions are met:                     #
#                                                                                                 #
# 1. Redistributions of source code must retain the above copyright notice, this                  #
#    list of conditions and the following disclaimer.                                             #
# 2. Redistributions in binary form must reproduce the above copyright notice,                    #
#    this list of conditions and the following disclaimer in the documentation                    #
#    and/or other materials provided with the distribution.                                       #
#                                                                                                 #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND                 #
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED                   #
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE                          #
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR                 #
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES                  #
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;                    #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND                     #
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT                      #
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS                   #
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                    #
#                                                                                                 #
# Author: Gianluca Frison, gianluca.frison (at) imtek.uni-freiburg.de                             #
#                                                                                                 #
###################################################################################################

from setuptools import setup, find_packages

import hpipm_python

setup(name='hpipm-python',
   version='0.2',
   description='Python interface to HPIPM',
   url='http://github.com/giaf/hpipm',
   author='Andrea Zanelli - Gianluca Frison',
   license='BSD-2',
   packages = find_packages(),
   zip_safe=False)



# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

import os
import numpy as np
from .acados_model import AcadosModel
from .acados_dims import AcadosSimDims
from .utils import get_acados_path, get_shared_lib_ext

class AcadosSimOptions:
    """
    class containing the solver options
    """
    def __init__(self):
        self.__integrator_type = 'ERK'
        self.__collocation_type = 'GAUSS_LEGENDRE'
        self.__Tsim = None
        # ints
        self.__sim_method_num_stages = 4
        self.__sim_method_num_steps = 1
        self.__sim_method_newton_iter = 3
        # doubles
        self.__sim_method_newton_tol = 0.0
        # bools
        self.__sens_forw = True
        self.__sens_adj = False
        self.__sens_algebraic = False
        self.__sens_hess = False
        self.__output_z = True
        self.__sim_method_jac_reuse = 0
        env = os.environ
        self.__ext_fun_compile_flags = '-O2' if 'ACADOS_EXT_FUN_COMPILE_FLAGS' not in env else env['ACADOS_EXT_FUN_COMPILE_FLAGS']
        self.__num_threads_in_batch_solve: int = 1

    @property
    def integrator_type(self):
        """Integrator type. Default: 'ERK'."""
        return self.__integrator_type

    @property
    def num_stages(self):
        """Number of stages in the integrator. Default: 1"""
        return self.__sim_method_num_stages

    @property
    def num_steps(self):
        """Number of steps in the integrator. Default: 1"""
        return self.__sim_method_num_steps

    @property
    def newton_iter(self):
        """Number of Newton iterations in simulation method. Default: 3"""
        return self.__sim_method_newton_iter

    @property
    def newton_tol(self):
        """
        Tolerance for Newton system solved in implicit integrator (IRK, GNSF).
        0.0 means this is not used and exactly newton_iter iterations are carried out.
        Default: 0.0
        """
        return self.__sim_method_newton_tol

    @property
    def sens_forw(self):
        """Boolean determining if forward sensitivities are computed. Default: True"""
        return self.__sens_forw

    @property
    def sens_adj(self):
        """Boolean determining if adjoint sensitivities are computed. Default: False"""
        return self.__sens_adj

    @property
    def sens_algebraic(self):
        """Boolean determining if sensitivities wrt algebraic variables are computed. Default: False"""
        return self.__sens_algebraic

    @property
    def sens_hess(self):
        """Boolean determining if hessians are computed. Default: False"""
        return self.__sens_hess

    @property
    def output_z(self):
        """Boolean determining if values for algebraic variables (corresponding to start of simulation interval) are computed. Default: True"""
        return self.__output_z

    @property
    def sim_method_jac_reuse(self):
        """Integer determining if jacobians are reused (0 or 1). Default: 0"""
        return self.__sim_method_jac_reuse

    @property
    def T(self):
        """Time horizon"""
        return self.__Tsim

    @property
    def collocation_type(self):
        """Collocation type: relevant for implicit integrators
        -- string in {'GAUSS_RADAU_IIA', 'GAUSS_LEGENDRE', 'EXPLICIT_RUNGE_KUTTA'}.

        Default: GAUSS_LEGENDRE
        """
        return self.__collocation_type

    @property
    def ext_fun_compile_flags(self):
        """
        String with compiler flags for external function compilation.
        Default: '-O2' if environment variable ACADOS_EXT_FUN_COMPILE_FLAGS is not set, else ACADOS_EXT_FUN_COMPILE_FLAGS is used as default.
        """
        return self.__ext_fun_compile_flags

    @property
    def num_threads_in_batch_solve(self):
        """
        Integer indicating how many threads should be used within the batch solve.
        If more than one thread should be used, the sim solver is compiled with openmp.
        Default: 1.
        """
        return self.__num_threads_in_batch_solve


    @ext_fun_compile_flags.setter
    def ext_fun_compile_flags(self, ext_fun_compile_flags):
        if isinstance(ext_fun_compile_flags, str):
            self.__ext_fun_compile_flags = ext_fun_compile_flags
        else:
            raise Exception('Invalid ext_fun_compile_flags, expected a string.\n')

    @integrator_type.setter
    def integrator_type(self, integrator_type):
        integrator_types = ('ERK', 'IRK', 'GNSF')
        if integrator_type in integrator_types:
            self.__integrator_type = integrator_type
        else:
            raise Exception('Invalid integrator_type value. Possible values are:\n\n' \
                    + ',\n'.join(integrator_types) + '.\n\nYou have: ' + integrator_type + '.\n\n')

    @collocation_type.setter
    def collocation_type(self, collocation_type):
        collocation_types = ('GAUSS_RADAU_IIA', 'GAUSS_LEGENDRE')
        if collocation_type in collocation_types:
            self.__collocation_type = collocation_type
        else:
            raise Exception('Invalid collocation_type value. Possible values are:\n\n' \
                    + ',\n'.join(collocation_types) + '.\n\nYou have: ' + collocation_type + '.\n\n')

    @T.setter
    def T(self, T):
        self.__Tsim = T

    @num_stages.setter
    def num_stages(self, num_stages):
        if isinstance(num_stages, int):
            self.__sim_method_num_stages = num_stages
        else:
            raise Exception('Invalid num_stages value. num_stages must be an integer.')

    @num_steps.setter
    def num_steps(self, num_steps):
        if isinstance(num_steps, int):
            self.__sim_method_num_steps = num_steps
        else:
            raise Exception('Invalid num_steps value. num_steps must be an integer.')

    @newton_iter.setter
    def newton_iter(self, newton_iter):
        if isinstance(newton_iter, int):
            self.__sim_method_newton_iter = newton_iter
        else:
            raise Exception('Invalid newton_iter value. newton_iter must be an integer.')

    @newton_tol.setter
    def newton_tol(self, newton_tol):
        if isinstance(newton_tol, float):
            self.__sim_method_newton_tol = newton_tol
        else:
            raise Exception('Invalid newton_tol value. newton_tol must be an float.')

    @sens_forw.setter
    def sens_forw(self, sens_forw):
        if sens_forw in (True, False):
            self.__sens_forw = sens_forw
        else:
            raise Exception('Invalid sens_forw value. sens_forw must be a Boolean.')

    @sens_adj.setter
    def sens_adj(self, sens_adj):
        if sens_adj in (True, False):
            self.__sens_adj = sens_adj
        else:
            raise Exception('Invalid sens_adj value. sens_adj must be a Boolean.')

    @sens_hess.setter
    def sens_hess(self, sens_hess):
        if sens_hess in (True, False):
            self.__sens_hess = sens_hess
        else:
            raise Exception('Invalid sens_hess value. sens_hess must be a Boolean.')

    @sens_algebraic.setter
    def sens_algebraic(self, sens_algebraic):
        if sens_algebraic in (True, False):
            self.__sens_algebraic = sens_algebraic
        else:
            raise Exception('Invalid sens_algebraic value. sens_algebraic must be a Boolean.')

    @output_z.setter
    def output_z(self, output_z):
        if output_z in (True, False):
            self.__output_z = output_z
        else:
            raise Exception('Invalid output_z value. output_z must be a Boolean.')

    @sim_method_jac_reuse.setter
    def sim_method_jac_reuse(self, sim_method_jac_reuse):
        if sim_method_jac_reuse in (0, 1):
            self.__sim_method_jac_reuse = sim_method_jac_reuse
        else:
            raise Exception('Invalid sim_method_jac_reuse value. sim_method_jac_reuse must be 0 or 1.')

    @num_threads_in_batch_solve.setter
    def num_threads_in_batch_solve(self, num_threads_in_batch_solve):
        if isinstance(num_threads_in_batch_solve, int) and num_threads_in_batch_solve > 0:
            self.__num_threads_in_batch_solve = num_threads_in_batch_solve
        else:
            raise Exception('Invalid num_threads_in_batch_solve value. num_threads_in_batch_solve must be a positive integer.')

class AcadosSim:
    """
    The class has the following properties that can be modified to formulate a specific simulation problem, see below:

    :param acados_path: string with the path to acados. It is used to generate the include and lib paths.

    - :py:attr:`dims` of type :py:class:`acados_template.acados_dims.AcadosSimDims` - are automatically detected from model
    - :py:attr:`model` of type :py:class:`acados_template.acados_model.AcadosModel`
    - :py:attr:`solver_options` of type :py:class:`acados_template.acados_sim.AcadosSimOptions`

    - :py:attr:`acados_include_path` (set automatically)
    - :py:attr:`shared_lib_ext` (set automatically)
    - :py:attr:`acados_lib_path` (set automatically)
    - :py:attr:`parameter_values` - used to initialize the parameters (can be changed)

    """
    def __init__(self, acados_path=''):
        if acados_path == '':
            acados_path = get_acados_path()
        self.dims = AcadosSimDims()
        """Dimension definitions, automatically detected from :py:attr:`model`. Type :py:class:`acados_template.acados_dims.AcadosSimDims`"""
        self.model = AcadosModel()
        """Model definitions, type :py:class:`acados_template.acados_model.AcadosModel`"""
        self.solver_options = AcadosSimOptions()
        """Solver Options, type :py:class:`acados_template.acados_sim.AcadosSimOptions`"""

        self.acados_include_path = os.path.join(acados_path, 'include').replace(os.sep, '/') # the replace part is important on Windows for CMake
        """Path to acados include directory (set automatically), type: `string`"""
        self.acados_lib_path = os.path.join(acados_path, 'lib').replace(os.sep, '/') # the replace part is important on Windows for CMake
        """Path to where acados library is located (set automatically), type: `string`"""

        self.code_export_directory = 'c_generated_code'
        """Path to where code will be exported. Default: `c_generated_code`."""
        self.shared_lib_ext = get_shared_lib_ext()

        # get cython paths
        from sysconfig import get_paths
        self.cython_include_dirs = [np.get_include(), get_paths()['include']]

        self.__parameter_values = np.array([])
        self.__problem_class = 'SIM'

    @property
    def parameter_values(self):
        """:math:`p` - initial values for parameter - can be updated"""
        return self.__parameter_values

    @parameter_values.setter
    def parameter_values(self, parameter_values):
        if isinstance(parameter_values, np.ndarray):
            self.__parameter_values = parameter_values
        else:
            raise Exception('Invalid parameter_values value. ' +
                            f'Expected numpy array, got {type(parameter_values)}.')

    def make_consistent(self):
        dims = self.dims
        model = self.model
        model.make_consistent(dims)

        if self.parameter_values.shape[0] != dims.np:
            raise Exception('inconsistent dimension np, regarding model.p and parameter_values.' + \
                f'\nGot np = {dims.np}, acados_sim.parameter_values.shape = {self.parameter_values.shape[0]}\n')

        # check required arguments are given
        if self.solver_options.T is None:
            raise Exception('acados_sim.solver_options.T is None, should be provided.')



###################################################################################################
#                                                                                                 #
# This file is part of HPIPM.                                                                     #
#                                                                                                 #
# HPIPM -- High-Performance Interior Point Method.                                                #
# Copyright (C) 2019 by Gianluca Frison.                                                          #
# Developed at IMTEK (University of Freiburg) under the supervision of Moritz Diehl.              #
# All rights reserved.                                                                            #
#                                                                                                 #
# The 2-Clause BSD License                                                                        #
#                                                                                                 #
# Redistribution and use in source and binary forms, with or without                              #
# modification, are permitted provided that the following conditions are met:                     #
#                                                                                                 #
# 1. Redistributions of source code must retain the above copyright notice, this                  #
#    list of conditions and the following disclaimer.                                             #
# 2. Redistributions in binary form must reproduce the above copyright notice,                    #
#    this list of conditions and the following disclaimer in the documentation                    #
#    and/or other materials provided with the distribution.                                       #
#                                                                                                 #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND                 #
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED                   #
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE                          #
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR                 #
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES                  #
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;                    #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND                     #
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT                      #
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS                   #
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                    #
#                                                                                                 #
# Author: Gianluca Frison, gianluca.frison (at) imtek.uni-freiburg.de                             #
#                                                                                                 #
###################################################################################################

from ctypes import *
import ctypes.util 
import numpy as np



class hpipm_ocp_qcqp_solver:


	def __init__(self, qp_dims, arg):

		# load hpipm shared library
		__hpipm   = CDLL('libhpipm.so')
		self.__hpipm = __hpipm

		# set up ipm workspace struct
		sizeof_d_ocp_qcqp_ipm_workspace = __hpipm.d_ocp_qcqp_ipm_ws_strsize()
		ipm_ws_struct = cast(create_string_buffer(sizeof_d_ocp_qcqp_ipm_workspace), c_void_p)
		self.ipm_ws_struct = ipm_ws_struct

		# allocate memory for ipm workspace 
		ipm_size = __hpipm.d_ocp_qcqp_ipm_ws_memsize(qp_dims.dim_struct, arg.arg_struct)
		ipm_ws_mem = cast(create_string_buffer(ipm_size), c_void_p)
		self.ipm_ws_mem = ipm_ws_mem

		# create C ws
		__hpipm.d_ocp_qcqp_ipm_ws_create(qp_dims.dim_struct, arg.arg_struct, ipm_ws_struct, ipm_ws_mem)

		self.arg = arg
		self.dim_struct = qp_dims.dim_struct
		

	def solve(self, qp, qp_sol):
		self.__hpipm.d_ocp_qcqp_ipm_solve(qp.qp_struct, qp_sol.qp_sol_struct, self.arg.arg_struct, self.ipm_ws_struct)


	def get(self, field):
		if((field=='stat')):
			# get iters
			iters = np.zeros((1,1), dtype=int);
			tmp = cast(iters.ctypes.data, POINTER(c_int))
			self.__hpipm.d_ocp_qcqp_ipm_get_iter(self.ipm_ws_struct, tmp)
			# get stat_m
			stat_m = np.zeros((1,1), dtype=int);
			tmp = cast(stat_m.ctypes.data, POINTER(c_int))
			self.__hpipm.d_ocp_qcqp_ipm_get_stat_m(self.ipm_ws_struct, tmp)
			# get stat pointer
			res = np.zeros((iters[0][0]+1, stat_m[0][0]));
			ptr = c_void_p()
			self.__hpipm.d_ocp_qcqp_ipm_get_stat(self.ipm_ws_struct, byref(ptr))
			tmp = cast(ptr, POINTER(c_double))
			for ii in range(iters[0][0]+1):
				for jj in range(stat_m[0][0]):
					res[ii][jj] = tmp[jj+ii*stat_m[0][0]]
					res[ii][jj] = tmp[jj+ii*stat_m[0][0]]
			return res
		elif((field=='status') | (field=='iter')):
			res = np.zeros((1,1), dtype=int);
			tmp = cast(res.ctypes.data, POINTER(c_int))
		elif((field=='max_res_stat') | (field=='max_res_eq') | (field=='max_res_ineq') | (field=='max_res_comp')):
			res = np.zeros((1,1));
			tmp = cast(res.ctypes.data, POINTER(c_double))
		else:
			raise NameError('hpipm_ocp_qcqp_solver.get: wrong field')
		field_name_b = field.encode('utf-8')
		self.__hpipm.d_ocp_qcqp_ipm_get(c_char_p(field_name_b), self.ipm_ws_struct, tmp)
		return res[0][0]







# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

from .acados_ocp_solver import AcadosOcpSolver
from .acados_ocp import AcadosOcp
from .acados_ocp_iterate import AcadosOcpFlattenedBatchIterate
from typing import Optional, List, Tuple, Sequence
from ctypes import (POINTER, c_int, c_void_p, cast, c_double, c_char_p)
import numpy as np
import time

class AcadosOcpBatchSolver():
    """
    Batch Integrator for parallel integration.

        :param sim: type :py:class:`~acados_template.acados_sim.AcadosOcp`
        :param N_batch: batch size, positive integer
        :param json_file: Default: 'acados_sim.json'
        :verbose: bool, default: True
    """

    __ocp_solvers : List[AcadosOcpSolver]

    def __init__(self, ocp: AcadosOcp, N_batch: int, json_file: str = 'acados_ocp.json', verbose: bool=True):

        if not isinstance(N_batch, int) or N_batch <= 0:
            raise Exception("AcadosOcpBatchSolver: argument N_batch should be a positive integer.")

        self.__N_batch = N_batch
        self.__ocp_solvers = [AcadosOcpSolver(ocp, json_file=json_file, build=n==0, generate=n==0, verbose=verbose) for n in range(self.N_batch)]

        self.__shared_lib = self.ocp_solvers[0].shared_lib
        self.__acados_lib = self.ocp_solvers[0].acados_lib
        self.__name = self.ocp_solvers[0].name
        self.__ocp_solvers_pointer = (c_void_p * self.N_batch)()

        for i in range(self.N_batch):
            self.__ocp_solvers_pointer[i] = self.ocp_solvers[i].capsule

        getattr(self.__shared_lib, f"{self.__name}_acados_batch_solve").argtypes = [POINTER(c_void_p), c_int]
        getattr(self.__shared_lib, f"{self.__name}_acados_batch_solve").restype = c_void_p

        getattr(self.__shared_lib, f"{self.__name}_acados_batch_eval_params_jac").argtypes = [POINTER(c_void_p), c_int]
        getattr(self.__shared_lib, f"{self.__name}_acados_batch_eval_params_jac").restype = c_void_p

        getattr(self.__shared_lib, f"{self.__name}_acados_batch_eval_solution_sens_adj_p").argtypes = [POINTER(c_void_p), c_char_p, c_int, POINTER(c_double), c_int, c_int]
        getattr(self.__shared_lib, f"{self.__name}_acados_batch_eval_solution_sens_adj_p").restype = c_void_p

        getattr(self.__shared_lib, f"{self.__name}_acados_batch_set_flat").argtypes = [POINTER(c_void_p), c_char_p, POINTER(c_double), c_int, c_int]
        getattr(self.__shared_lib, f"{self.__name}_acados_batch_set_flat").restype = c_void_p

        getattr(self.__shared_lib, f"{self.__name}_acados_batch_get_flat").argtypes = [POINTER(c_void_p), c_char_p, POINTER(c_double), c_int, c_int]
        getattr(self.__shared_lib, f"{self.__name}_acados_batch_get_flat").restype = c_void_p

        if self.ocp_solvers[0].acados_lib_uses_omp:
            msg = "Note: Please make sure that the acados shared library is compiled with the number of threads set to 1,\n"
        else:
            msg = "Warning: Please compile the acados shared library with openmp and the number of threads set to 1,\n"

        msg += "i.e. with the flags -DACADOS_WITH_OPENMP=ON -DACADOS_NUM_THREADS=1.\n" + \
                   "See https://github.com/acados/acados/pull/1089 for more details."
        print(msg)


    @property
    def ocp_solvers(self):
        """List of AcadosOcpSolvers."""
        return self.__ocp_solvers


    @property
    def N_batch(self):
        """Batch size."""
        return self.__N_batch


    def solve(self):
        """
        Call solve for all `N_batch` solvers.
        """
        getattr(self.__shared_lib, f"{self.__name}_acados_batch_solve")(self.__ocp_solvers_pointer, self.__N_batch)



    def eval_adjoint_solution_sensitivity(self,
                                          seed_x: Optional[Sequence[Tuple[int, np.ndarray]]],
                                          seed_u: Optional[Sequence[Tuple[int, np.ndarray]]],
                                          with_respect_to: str = "p_global",
                                          sanity_checks: bool = True,
                                          ) -> np.ndarray:
        """
        Evaluate the adjoint sensitivity of the solution with respect to the parameters.
            :param seed_x : Sequence of tuples of the form (stage: int, seed_vec: np.ndarray).
                    The stage is the stage at which the seed_vec is applied, and seed_vec is the seed for the states at that stage with shape (N_batch, nx, n_seeds)
            :param seed_u : Sequence of tuples of the form (stage: int, seed_vec: np.ndarray).
                    The stage is the stage at which the seed_vec is applied, and seed_vec is the seed for the controls at that stage with shape (N_batch, nu, n_seeds)
            :param with_respect_to : string in ["p_global"]
            :param sanity_checks : bool - whether to perform sanity checks, turn off for minimal overhead, default: True
            :returns : np.ndarray of shape (N_batch, n_seeds, np_global)
        """

        if seed_x is None:
            seed_x = []
        elif not isinstance(seed_x, Sequence):
            raise Exception(f"seed_x should be a Sequence, got {type(seed_x)}")

        if seed_u is None:
            seed_u = []
        elif not isinstance(seed_u, Sequence):
            raise Exception(f"seed_u should be a Sequence, got {type(seed_u)}")

        if len(seed_x) == 0 and len(seed_u) == 0:
            raise Exception("seed_x and seed_u cannot both be empty.")

        if len(seed_x) > 0:
            if not isinstance(seed_x[0], tuple) or len(seed_x[0]) != 2:
                raise Exception(f"seed_x[0] should be tuple of length 2, got seed_x[0] {seed_x[0]}")
            s = seed_x[0][1]
            if not isinstance(s, np.ndarray):
                raise Exception(f"seed_x[0][1] should be np.ndarray, got {type(s)}")
            n_seeds = seed_x[0][1].shape[2]

        if len(seed_u) > 0:
            if not isinstance(seed_u[0], tuple) or len(seed_u[0]) != 2:
                raise Exception(f"seed_u[0] should be tuple of length 2, got seed_u[0] {seed_u[0]}")
            s = seed_u[0][1]
            if not isinstance(s, np.ndarray):
                raise Exception(f"seed_u[0][1] should be np.ndarray, got {type(s)}")
            n_seeds = seed_u[0][1].shape[2]

        if sanity_checks:
            N_horizon = self.__ocp_solvers[0].acados_ocp.solver_options.N_horizon

            for n in range(self.N_batch):
                self.__ocp_solvers[n]._sanity_check_solution_sensitivities()

            nx = self.__acados_lib.ocp_nlp_dims_get_from_attr(
                self.__ocp_solvers[0].nlp_config, self.__ocp_solvers[0].nlp_dims, self.__ocp_solvers[0].nlp_out, 0, "x".encode('utf-8'))
            nu = self.__acados_lib.ocp_nlp_dims_get_from_attr(
                self.__ocp_solvers[0].nlp_config, self.__ocp_solvers[0].nlp_dims, self.__ocp_solvers[0].nlp_out, 0, "u".encode('utf-8'))

            # check seeds
            for seed, name, dim in [(seed_x, "seed_x", nx,), (seed_u, "seed_u", nu)]:
                for stage, seed_stage in seed:
                    if not isinstance(stage, int) or stage < 0 or stage > N_horizon:
                        raise Exception(f"AcadosOcpBatchSolver.eval_adjoint_solution_sensitivity(): stage {stage} for {name} is not valid.")
                    if not isinstance(seed_stage, np.ndarray):
                        raise Exception(f"{name} for stage {stage} should be np.ndarray, got {type(seed_stage)}")
                    if seed_stage.shape != (self.N_batch, dim, n_seeds):
                        raise Exception(f"{name} for stage {stage} should have shape (N_batch, dim, n_seeds) = ({self.N_batch}, {dim}, {n_seeds}), got {seed_stage.shape}.")

        if with_respect_to == "p_global":
            field = "p_global".encode('utf-8')

            np_global = self.__acados_lib.ocp_nlp_dims_get_from_attr(
                self.__ocp_solvers[0].nlp_config, self.__ocp_solvers[0].nlp_dims, self.__ocp_solvers[0].nlp_out, 0, field)

            # compute jacobian wrt params
            t0 = time.time()
            getattr(self.__shared_lib, f"{self.__name}_acados_batch_eval_params_jac")(self.__ocp_solvers_pointer, self.__N_batch)
            self.time_solution_sens_lin = time.time() - t0

            t1 = time.time()

            grad_p = np.zeros((self.N_batch, n_seeds, np_global), order="C", dtype=np.float64)
            offset = n_seeds*np_global

            for i_seed in range(n_seeds):
                # set seed:
                self.reset_sens_out()

                # TODO this could be a batch_set
                for m in range(self.N_batch):
                    for (stage, sx) in seed_x:
                        self.ocp_solvers[m].set(stage, 'sens_x', sx[m, :, i_seed])
                    for (stage, su) in seed_u:
                        self.ocp_solvers[m].set(stage, 'sens_u', su[m, :, i_seed])

                c_grad_p = cast(grad_p[0, i_seed].ctypes.data, POINTER(c_double))

                # solve adjoint sensitivities
                getattr(self.__shared_lib, f"{self.__name}_acados_batch_eval_solution_sens_adj_p")(
                    self.__ocp_solvers_pointer, field, 0, c_grad_p, offset, self.__N_batch)

            self.time_solution_sens_solve = time.time() - t1

            return grad_p
        else:
            raise NotImplementedError(f"with_respect_to {with_respect_to} not implemented.")


    def reset_sens_out(self, ):
        # TODO batch this
        for n in range(self.N_batch):
            self.__acados_lib.ocp_nlp_out_set_values_to_zero(self.__ocp_solvers[n].nlp_config, self.__ocp_solvers[n].nlp_dims, self.__ocp_solvers[n].sens_out)



    def set_flat(self, field_: str, value_: np.ndarray) -> None:
        """
        Set concatenation solver initialization for all `N_batch` solvers.

            :param field_: string in ['x', 'u', 'z', 'pi', 'lam', 'sl', 'su', 'p']
            :param value_: np.array of shape (N_batch, n_field_total)
        """

        field = field_.encode('utf-8')
        if field_ not in ['x', 'u', 'z', 'pi', 'lam', 'sl', 'su', 'p']:
            raise Exception(f'AcadosOcpSolver.get_flat(field={field_}): \'{field_}\' is an invalid argument.')

        dim = self.ocp_solvers[0].get_dim_flat(field_)

        if value_.shape != (self.N_batch, dim):
            raise Exception(f'AcadosOcpBatchSolver.set_flat(field={field_}, value): value has wrong shape, expected ({self.N_batch}, {dim}), got {value_.shape}.')

        value_ = value_.reshape((-1,), order='C')
        N_data = value_.shape[0]

        value_ = value_.astype(float)
        value_data = cast(value_.ctypes.data, POINTER(c_double))

        getattr(self.__shared_lib, f"{self.__name}_acados_batch_set_flat")(self.__ocp_solvers_pointer, field, value_data, N_data, self.__N_batch)


    def get_flat(self, field_: str) -> np.ndarray:
        """
        Get concatenation of all stages of last solution of the solver.

            :param field: string in ['x', 'u', 'z', 'pi', 'lam', 'sl', 'su', 'p']
            :returns: numpy array of shape (N_batch, n_field_total)
        """
        if field_ not in ['x', 'u', 'z', 'pi', 'lam', 'sl', 'su', 'p']:
            raise Exception(f'AcadosOcpSolver.get_flat(field={field_}): \'{field_}\' is an invalid argument.')

        field = field_.encode('utf-8')

        dim = self.ocp_solvers[0].get_dim_flat(field_)

        out = np.ascontiguousarray(np.zeros((self.N_batch, dim,)), dtype=np.float64)
        out_data = cast(out.ctypes.data, POINTER(c_double))

        getattr(self.__shared_lib, f"{self.__name}_acados_batch_get_flat")(self.__ocp_solvers_pointer, field, out_data, self.N_batch*dim, self.__N_batch)

        return out


    def store_iterate_to_flat_obj(self) -> AcadosOcpFlattenedBatchIterate:
        """
        Returns the current iterate of the OCP solvers as an AcadosOcpFlattenedBatchIterate.
        """
        return AcadosOcpFlattenedBatchIterate(x = self.get_flat("x"),
                                              u = self.get_flat("u"),
                                              z = self.get_flat("z"),
                                              sl = self.get_flat("sl"),
                                              su = self.get_flat("su"),
                                              pi = self.get_flat("pi"),
                                              lam = self.get_flat("lam"),
                                              N_batch=self.N_batch)

    def load_iterate_from_flat_obj(self, iterate: AcadosOcpFlattenedBatchIterate) -> None:
        """
        Loads the provided iterate into the OCP solvers.
        Note: The iterate object does not contain the the parameters.
        """

        if self.N_batch != iterate.N_batch:
            raise Exception(f"Wrong batch dimension. Expected {self.N_batch}, got {iterate.N_batch}")

        self.set_flat("x", iterate.x)
        self.set_flat("u", iterate.u)
        self.set_flat("z", iterate.z)
        self.set_flat("sl", iterate.sl)
        self.set_flat("su", iterate.su)
        self.set_flat("pi", iterate.pi)
        self.set_flat("lam", iterate.lam)



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

from ..multiple_shooting import MultipleShooting
from ..sampling_method import SamplingMethod, UniformGrid
from ..solution import OcpSolution
from ..freetime import FreeTime
from ..casadi_helpers import DM2numpy, reshape_number, linear_coeffs
from collections import OrderedDict
from casadi import SX, Sparsity, MX, vcat, veccat, symvar, substitute, sparsify, DM, Opti, is_linear, vertcat, depends_on, jacobian, linear_coeff, quadratic_coeff, mtimes, pinv, evalf, Function, vvcat, inf, sum1, sum2, diag
import casadi
import numpy as np
import scipy
import casadi as ca

INF = 1e5

def legit_J(J):
    """
    Checks if J, a pre-multiplier for states and control, is of legitimate structure
    J must a slice of a permuted unit matrix.
    """
    try:
        J = evalf(J)
    except:
        return False
    if not np.all(np.array(J.nonzeros())==1): # All nonzeros must be one
        return False
    # Each row must contain exactly 1 nonzero
    if not np.all(np.array(sum2(J))==1):
        return False
    # Each column must contain at most 1 nonzero
    if not np.all(np.array(sum1(J))<=1):
        return False
    return True


def check_Js(J):
    """
    Checks if J, a pre-multiplier for slacks, is of legitimate structure
    Empty rows are allowed
    """
    try:
        J = evalf(J)
    except:
        raise Exception("Slack error")
    assert np.all(np.array(J.nonzeros())==1), "All nonzeros must be 1"
    # Check if slice of permutation of unit matrix
    assert np.all(np.array(sum2(J))<=1), "Each constraint can only depend on one slack at most"
    assert np.all(np.array(sum1(J))<=1), "Each constraint must depend on a unique slack, if any"

def fill_in(var, expr,value, array, k):
    value = evalf(value)
    J, r = linear_coeffs(var,expr)
    J = evalf(J)
    r = evalf(r)
    assert r.is_zero()
    check_Js(J)
    expr = reshape_number(var, expr)
    if J.sparsity().get_col():
        array[J.sparsity().get_col(), k] = value[J.row()]
def fill_in_array(var, expr, value, array, numeric = True):
    if numeric:
        value = evalf(value)
    J, r = linear_coeffs(var[:,0],expr)
    J = evalf(J)
    r = evalf(r)
    assert r.is_zero()
    check_Js(J)
    expr = reshape_number(var, expr)
    if J.sparsity().get_col():
        array[J.sparsity().get_col(), :] = value[J.row(), :]

class ExternalMethod:
    def __init__(self,supported=None,N=20,grid=UniformGrid(),linesearch=True,expand=False,**args):
        self.N = N
        self.args = args
        self.grid = grid
        self.T = None
        self.linesearch = linesearch
        self.expand = expand
        self.supported = {} if supported is None else supported
        self.free_time = False

    def inherit(self, parent):
        pass

    def fill_placeholders_t0(self, phase, stage, expr, *args):
        if phase==1:
            if isinstance(stage._t0,FreeTime):
                return stage.at_t0(self.t)
            else:
                return stage._t0
        return

    def fill_placeholders_T(self, phase, stage, expr, *args):
        if phase==1:
            if isinstance(stage._T,FreeTime):
                self.free_time = True
                if "free_T" in self.supported:
                    # Keep placeholder symbol
                    return
                else:
                    T = stage.register_state(MX.sym('T'))
                    stage.set_der(T, 0)
                    self.T = T
                    stage.set_initial(T, stage._T.T_init)
                    return stage.at_tf(T)
            else:
                self.T = stage._T
                return stage._T
        return

    def fill_placeholders_t(self, phase, stage, expr, *args):
        if phase==1:
            if self.t_state:
                t = stage.state()
                stage.set_next(t, t+ stage.DT) if stage._state_next else stage.set_der(t, 1)
                self.t = t
                if not isinstance(stage._t0,FreeTime):
                    stage.subject_to(stage.at_t0(self.t)==stage._t0)
                return t
            else:
                self.t = MX(1,1)
                return
        return

    def transcribe_placeholders(self, phase, stage, placeholders):
        """
        Transcription is the process of going from a continuous-time OCP to an NLP
        """
        return stage._transcribe_placeholders(phase, self, placeholders)


    def fill_placeholders_sum_control(self, phase, stage, expr, *args):
        if phase==1:
            return expr

    def fill_placeholders_at_t0(self, phase, stage, expr, *args):
        if phase==1: return
        ret = {}
        ks = [stage.x,stage.u]
        vs = [self.X_gist[0],self.U_gist[0]]
        if not self.t_state:
            ks += [stage.t]
            vs += [self.control_grid[0]]
        ret["normal"] = substitute([expr],ks,vs)[0]
        ret["expose"] = expr
        return ret

    def fill_placeholders_at_tf(self, phase, stage, expr, *args):
        if phase==1: return
        ret = {}
        ks = [stage.x,stage.u]
        vs = [self.X_gist[-1],self.U_gist[-1]]
        if not self.t_state:
            ks += [stage.t]
            vs += [self.control_grid[-1]]
        ret["normal"] = substitute([expr],ks,vs)[0]
        ret["expose"] = expr
        return ret

    def fill_placeholders_DT(self, phase, stage, expr, *args):
        return None

    def fill_placeholders_DT_control(self, phase, stage, expr, *args):
        return None

    def main_transcribe(self, stage, phase=1, **kwargs):
        pass

    def transcribe(self, stage, phase=1, **kwargs):

        if phase==0:

            def recursive_depends(expr,t):
                expr = MX(expr)
                if depends_on(expr, t):
                    return True
                if expr in stage._placeholders:
                    if recursive_depends(stage._placeholders[expr][1],t):
                        return True
                else:
                    if expr.is_symbolic():
                        return depends_on(expr, t)
                    for s in symvar(expr):
                        if recursive_depends(s, t):
                            return True
                return False 
                  
            # Do we need to introduce a helper state for t?
            f = stage._hybrid()[0]
            if not stage._state_next:
                if f.sparsity_in('t').nnz()>0:
                    self.t_state = True
                    return

            # Occurs in lagrange?
            obj = MX(stage._objective)
            for e in symvar(obj):
                if e.name()=='r_sum_control' and recursive_depends(e,stage.t):
                    self.t_state = True
                    return

            if isinstance(stage._t0, FreeTime):
                self.t_state = True
                return

            if isinstance(stage._T, FreeTime) or isinstance(stage._t0, FreeTime):
                for c,_,_ in stage._constraints["control"]+stage._constraints["integrator"]:
                    if recursive_depends(c,stage.t):
                        self.t_state = True
                        return
            
            self.t_state = False
            return

        if phase==1:

            self.p_global = stage.parameters['']
            self.p_local = stage.parameters['control']+stage.parameters['control+']

            self.p_global_cat = ca.vvcat(self.p_global)
            self.p_local_cat = ca.vvcat(self.p_local)


            self.p_global_value = DM.zeros(ca.vvcat(self.p_global).sparsity())
            self.p_local_value = DM.zeros(ca.vvcat(self.p_local).numel(), self.N+1)

            self.transcribe_phase1(stage, **kwargs)
        else:
            self.transcribe_phase2(stage, **kwargs)

    def set_value(self, stage, master, parameter, value):
            value = evalf(value)
            if parameter in self.p_global:
                p = ca.vec(parameter)
                fill_in(p, self.p_global_cat, value, self.p_global_value,0)
            elif parameter in self.p_local:
                p = ca.vec(parameter)
                # for k in list(range(self.N)):
                if p.numel()==value.numel():
                    if(value.shape[1] == p.shape[0]):
                        value = value.T
                    value = ca.repmat(value,  1, self.N+1)
                if p.numel()*(self.N)==value.numel() or p.numel()*(self.N+1)== value.numel():
                    if(value.shape[1] == p.shape[0]):
                        value = value.T
                assert(value.shape[0] == p.shape[0])
                if p.numel()*(self.N)== value.numel():
                    value = ca.horzcat(value, value[:,-1])
                fill_in_array(p, self.p_local_cat, value, self.p_local_value)
            else:
                raise Exception("Parameter not found")

    @property
    def gist(self):
        """Obtain an expression packing all information needed to obtain value/sample

        The composition of this array may vary between rockit versions

        Returns
        -------
        :obj:`~casadi.MX` column vector

        """
        ret = []
        for e,_,_ in self.gist_parts:
            if isinstance(e, list):
                ret.append(vcat(e))
            else:
                ret.append(e)
        return vcat(ret)
    

    def eval(self, stage, expr):
        placeholders = stage.placeholders_transcribed
        expr = placeholders(expr,max_phase=1)
        ks = []
        vs = []
        for g,s,mode in self.gist_parts:
            if mode=="global":
                ks.append(s)
                vs.append(g)
        ret = substitute([expr],ks,vs)[0]
        return ret

    def eval_at_control(self, stage, expr, k):
        placeholders = stage.placeholders_transcribed
        expr = placeholders(expr,max_phase=1)
        ks = []
        vs = []
        for g,s,mode in self.gist_parts:
           ks.append(s)
           if isinstance(g, list):
               g = g[min(k,len(g)-1)]
           vs.append(g)
        if not self.t_state:
            ks += [stage.t]
            vs += [self.control_grid[k]]
        ret = substitute([expr],ks,vs)[0]
        return ret


class SolWrapper:
    def __init__(self, method, x, u):
        self.method = method
        self.x = x
        self.u = u

    def value(self, expr, *args,**kwargs):
        placeholders = self.method.stage.placeholders_transcribed
        ret = evalf(substitute([placeholders(expr)],[self.method.gist],[vertcat(self.x, self.u)])[0])
        return ret.toarray(simplify=True)



#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#
#   Author: Jonathan Frey: jonathanpaulfrey(at)gmail.com

from casadi import *
from .determine_input_nonlinearity_function import determine_input_nonlinearity_function
from .check_reformulation import check_reformulation


def reformulate_with_invertible_E_mat(gnsf, model, print_info):
    ## Description
    # this function checks that the necessary condition to apply the gnsf
    # structure exploiting integrator to a model, namely that the matrices E11,
    # E22 are invertible holds.
    # if this is not the case, it will make these matrices invertible and add:
    # corresponding terms, to the term C * phi, such that the obtained model is
    # still equivalent

    # check invertibility of E11, E22 and reformulate if needed:
    ind_11 = range(gnsf["nx1"])
    ind_22 = range(gnsf["nx1"], gnsf["nx1"] + gnsf["nz1"])

    if print_info:
        print(" ")
        print("----------------------------------------------------")
        print("checking rank of E11 and E22")
        print("----------------------------------------------------")
    ## check if E11, E22 are invertible:
    z_check = False
    if gnsf["nz1"] > 0:
        z_check = (
            np.linalg.matrix_rank(gnsf["E"][np.ix_(ind_22, ind_22)]) != gnsf["nz1"]
        )

    if (
        np.linalg.matrix_rank(gnsf["E"][np.ix_(ind_11, ind_11)]) != gnsf["nx1"]
        or z_check
    ):
        # print warning (always)
        print(f"the rank of E11 or E22 is not full after the reformulation")
        print("")
        print(
            f"the script will try to reformulate the model with an invertible matrix instead"
        )
        print(
            f"NOTE: this feature is based on a heuristic, it should be used with care!!!"
        )

        ## load models
        xdot = gnsf["xdot"]
        z = gnsf["z"]

        # # GNSF
        # get dimensions
        nx1 = gnsf["nx1"]
        x1dot = xdot[range(nx1)]

        k = vertcat(x1dot, z)
        for i in [1, 2]:
            if i == 1:
                ind = range(gnsf["nx1"])
            else:
                ind = range(gnsf["nx1"], gnsf["nx1"] + gnsf["nz1"])
            mat = gnsf["E"][np.ix_(ind, ind)]
            import pdb

            pdb.set_trace()
            while np.linalg.matrix_rank(mat) < len(ind):
                # import pdb; pdb.set_trace()
                if print_info:
                    print(" ")
                    print(f"the rank of E", str(i), str(i), " is not full")
                    print(
                        f"the algorithm will try to reformulate the model with an invertible matrix instead"
                    )
                    print(
                        f"NOTE: this feature is not super stable and might need more testing!!!!!!"
                    )
                for sub_max in ind:
                    sub_ind = range(min(ind), sub_max)
                    # regard the submatrix mat(sub_ind, sub_ind)
                    sub_mat = gnsf["E"][sub_ind, sub_ind]
                    if np.linalg.matrix_rank(sub_mat) < len(sub_ind):
                        # reformulate the model by adding a 1 to last diagonal
                        # element and changing rhs respectively.
                        gnsf["E"][sub_max, sub_max] = gnsf["E"][sub_max, sub_max] + 1
                        # this means adding the term 1 * k(sub_max) to the sub_max
                        # row of the l.h.s
                        if len(np.nonzero(gnsf["C"][sub_max, :])[0]) == 0:
                            # if isempty(find(gnsf['C'](sub_max,:), 1)):
                            # add new nonlinearity entry
                            gnsf["C"][sub_max, gnsf["n_out"] + 1] = 1
                            gnsf["phi_expr"] = vertcat(gnsf["phi_expr"], k[sub_max])
                        else:
                            ind_f = np.nonzero(gnsf["C"][sub_max, :])[0]
                            if len(ind_f) != 1:
                                raise Exception("C is assumed to be a selection matrix")
                            else:
                                ind_f = ind_f[0]
                            # add term to corresponding nonlinearity entry
                            # note: herbey we assume that C is a selection matrix,
                            # i.e. gnsf['phi_expr'](ind_f) is only entering one equation

                            gnsf["phi_expr"][ind_f] = (
                                gnsf["phi_expr"][ind_f]
                                + k[sub_max] / gnsf["C"][sub_max, ind_f]
                            )
                            gnsf = determine_input_nonlinearity_function(gnsf)
                            check_reformulation(model, gnsf, print_info)
        print("successfully reformulated the model with invertible matrices E11, E22")
    else:
        if print_info:
            print(" ")
            print(
                "the rank of both E11 and E22 is naturally full after the reformulation "
            )
            print("==>  model reformulation finished")
            print(" ")
    if (gnsf['nx2'] > 0 or gnsf['nz2'] > 0) and det(gnsf["E_LO"]) == 0:
        print(
            "_______________________________________________________________________________________________________"
        )
        print(" ")
        print("TAKE CARE ")
        print("E_LO matrix is NOT regular after automatic transcription!")
        print("->> this means the model CANNOT be used with the gnsf integrator")
        print(
            "->> it probably means that one entry (of xdot or z) that was moved to the linear output type system"
        )
        print("    does not appear in the model at all (zero column in E_LO)")
        print(" OR: the columns of E_LO are linearly dependent ")
        print(" ")
        print(
            " SOLUTIONs: a) go through your model & check equations the method wanted to move to LOS"
        )
        print("            b) deactivate the detect_LOS option")
        print(
            "_______________________________________________________________________________________________________"
        )
    return gnsf
